"""AIMS output parser, taken from ASE with modifications."""

from __future__ import annotations

import numpy as np

from typing import Any, Sequence

from pyfhiaims.geometry.atom import FHIAimsAtom
from pyfhiaims.geometry.geometry import AimsGeometry
from pyfhiaims.output_parser.aims_out_section import AimsOutSection, AimsParseError
from pyfhiaims.output_parser.aims_out_header_section import AimsOutHeaderSection
from pyfhiaims.utils.typecast import Vector3D, Matrix3D, to_vector3d

PARSE_CALC_KEYS = {
    "geometry_start": [
        "Atomic structure (and velocities) as used in the preceding time step",
        "Updated atomic structure",
        "Atomic structure that was used in the preceding time step of the wrapper",
    ],
    "geometry_end": ['Writing the current geometry to file "geometry.in.next_step"'],
    "forces": ["Total atomic forces"],
    "stresses": ["Per atom stress (eV) used for heat flux calculation"],
    "stress": [
        "Analytical stress tensor - Symmetrized",
        "Numerical stress tensor",
    ],
    "is_metallic": ["material is metallic within the approximate finite broadening function (occupation_type)"],
    "total_energy_uncorrected": ["Total energy uncorrected"],
    "total_energy_corrected": ["Total energy corrected"],
    "free_energy": ["| Electronic free energy"],
    "number_of_iterations": ["| Number of self-consistency cycles"],
    "magnetic_moment": ["N_up - N_down"],
    "fermi_energy": ["| Chemical potential (Fermi level)"],
    "dipole": ["Total dipole moment [eAng]"],
    "dielectric_tensor": ["PARSE DFPT_dielectric_tensor"],
    "polarization": ["| Cartesian Polarization"],
    "homo": ["Highest occupied state"],
    "lumo": ["Lowest unoccupied state"],
    "gap": ["verall HOMO-LUMO gap"],
    "direct_gap": ["Smallest direct gap"],
    "hirshfeld_charge": ["Hirshfeld charge"],
    "hirshfeld_volume": ["Hirshfeld volume"],
    "hirshfeld_dipole": ["Hirshfeld dipole vector"],
    "mulliken_charges": ["Summary of the per-atom charge analysis"],
    "mulliken_spins": ["Summary of the per-atom spin analysis"],
}


class AimsOutCalcSection(AimsOutSection):
    """A part of the aims.out file corresponding to a single structure."""

    def __init__(self, lines: list[str], header: AimsOutHeaderSection) -> None:
        """Construct the AimsOutCalcChunk.

        Args:
            lines (list[str]): The lines used for the structure
            header (.AimsOutHeaderChunk):  A summary of the relevant information from
                the aims.out header
        """
        super().__init__(lines)
        self._header = header
        self.set_prop_line_relations(PARSE_CALC_KEYS)

    def _parse_geometry(self) -> AimsGeometry:
        """Parse a structure object from the file.

        For the given section of the aims output file generate the
        calculated structure.

        Returns:
            The structure or molecule for the calculation
        """
        symbols, positions, velocities, lattice = self._parse_lattice_atom_pos()
        atoms = []
        for aa, atom in enumerate(self.initial_geometry.atoms):
            atoms.append(
                FHIAimsAtom(
                    symbol=symbols[aa],
                    position=positions[aa],
                    velocity=velocities.get(aa),
                    constraints=atom.constraints,
                    constraint_region=atom.constraint_region,
                    magnetic_response=atom.magnetic_response,
                    magnetic_moment=atom.magnetic_moment,
                    nuclear_spin=atom.nuclear_spin,
                    isotope=atom.isotope,
                    is_empty=atom.is_empty,
                    is_pseudocore=atom.is_pseudocore,
                    RT_TDDFT_initial_velocity=atom.RT_TDDFT_initial_velocity,
                )
            )
        return AimsGeometry(
            atoms=atoms,
            lattice_vectors=lattice,
            lattice_constraints=self.initial_geometry.lattice_constraints,
            hessian_block=self.initial_geometry.hessian_block,
            hessian_block_lv=self.initial_geometry.hessian_block_lv,
            hessian_block_lv_atoms=self.initial_geometry.hessian_block_lv_atoms,
            hessian_file=self.initial_geometry.hessian_file,
            trust_radius=self.initial_geometry.trust_radius,
            symmetry_n_params=self.initial_geometry.symmetry_n_params,
            symmetry_params=self.initial_geometry.symmetry_params,
            symmetry_lv=self.initial_geometry.symmetry_lv,
            symmetry_frac=self.initial_geometry.symmetry_frac,
            symmetry_frac_change_threshold=self.initial_geometry.symmetry_frac_change_threshold,
            homogeneous_field=self.initial_geometry.homogeneous_field,
            multipole=self.initial_geometry.multipole,
            esp_constraint=self.initial_geometry.esp_constraint,
            verbatim_writeout=self.initial_geometry.verbatim_writeout,
            calculate_friction=self.initial_geometry.calculate_friction,
        )

    def _parse_lattice_atom_pos(
        self,
    ) -> tuple[list[str], list[Vector3D], dict[int, Vector3D], Matrix3D | None]:
        """Parse the lattice and atomic positions of the structure.

        Returns:
            list[str]: The species symbols for the atoms in the structure
            list[Vector3D]: The Cartesian coordinates of the atoms
            list[Vector3D]: The velocities of the atoms
            Lattice or None: The Lattice for the structure
        """
        
        try:
            line_start = self.get_line_inds("geometry_start")[-1]
        except IndexError:
            return (
                self.initial_geometry.symbols, 
                self.initial_geometry.positions, 
                {aa: vel for aa, vel in enumerate(self.initial_geometry.velocities)},
                self.initial_geometry.lattice_vectors
            )


        lattice_vectors: list[list[float]] = []
        velocities: dict[int, Vector3D] = {}
        symbols: list[str] = []
        coords: list[Vector3D] = []

        line_start += 1
        try:
            line_end = self.get_line_inds("geometry_end")[-1]
        except IndexError:
            line_end = len(self.lines)

        for line in self.lines[line_start:line_end]:
            if "lattice_vector   " in line:
                lattice_vectors.append([float(inp) for inp in line.split()[1:]])
            elif "atom " in line:
                line_split = line.split()
                symbols.append(line_split[4])
                coords.append(to_vector3d([float(x) for x in line_split[1:4]]))
            elif "velocity " in line:
                velocities[len(symbols) - 1] = to_vector3d([float(x) for x in line.split()[1:4]])

        lattice = np.array(lattice_vectors) if len(lattice_vectors) == 3 else None
        return symbols, coords, velocities, lattice

    @property
    def symbols(self) -> list[str]:
        """The list of atomic symbols for all atoms in the structure."""
        return self.geometry.symbols

    @property
    def positions(self) -> list[str]:
        """The list of atomic positions for all atoms in the structure."""
        return self.geometry.positions

    @property
    def velocities(self) -> list[str]:
        """The list of atomic velocities for all atoms in the structure."""
        return self.geometry.velocities
    
    @property
    def lattice(self) -> Matrix3D | None:
        """The lattice vectors for the geometry."""
        return self.geometry.lattice_vectors

    @property
    def forces(self) -> np.array[Vector3D] | None:
        """The forces from the aims.out file."""
        if "forces" in self._cache:
            return self._cache["forces"]
        
        try:
            line_start = self.get_line_inds("forces")[-1]
        except IndexError:
            self._cache["forces"] = None
            return None

        line_start += 1
        self._cache["forces"] = np.array(
            [[float(inp) for inp in line.split()[-3:]] for line in self.lines[line_start : line_start + self.n_atoms]]
        )
        return self._cache["forces"]

    @property
    def stresses(self) -> np.array[Matrix3D] | None:
        """The stresses from the aims.out file and convert to kbar."""
        if "stresses" in self._cache:
            return self._cache["stresses"]
        
        try:
            line_start = self.get_line_inds("stresses")[-1]
        except IndexError:
            self._cache["stresses"] = None
            return None

        line_start += 3
        stresses = []
        for line in self.lines[line_start : line_start + self.n_atoms]:
            xx, yy, zz, xy, xz, yz = (float(d) for d in line.split()[2:8])
            stresses.append(np.array([[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]]))

        self._cache["stresses"] = np.array(stresses)
        return  self._cache["stresses"]

    @property
    def stress(self) -> Matrix3D | None:
        """The stress from the aims.out file and convert to kbar."""
        if "stress" in self._cache:
            return self._cache["stress"]
        
        try:
            line_start = self.get_line_inds("stress")[-1]
        except IndexError:
            self._cache["stress"] = None
            return None
        stress = [[float(inp) for inp in line.split()[2:5]] for line in self.lines[line_start + 5 : line_start + 8]]
        self._cache["stress"] = np.array(stress)
        return self._cache["stress"]

    @property
    def is_metallic(self) -> bool:
        """Is the system is metallic."""
        return len(self.get_line_inds("is_metallic")) > 0

    @property
    def total_corrected_energy(self) -> float:
        """The total corrected energy from the aims.out file."""
        if "total_energy_corrected" in self._cache:
            return self._cache["total_energy_corrected"]
        
        try:
            line_ind = self.get_line_inds("total_energy_corrected")[-1]
        except IndexError:
            raise AimsParseError("No energy is associated with the structure.")

        self._cache["total_energy_corrected"] = float(self.lines[line_ind].split()[5])
        return self._cache["total_energy_corrected"]

    @property
    def total_uncorrected_energy(self) -> float:
        """The total uncorrected energy from the aims.out file."""
        if "total_energy_uncorrected" in self._cache:
            return self._cache["total_energy_uncorrected"]
        
        try:
            line_ind = self.get_line_inds("total_energy_uncorrected")[-1]
        except IndexError:
            raise AimsParseError("No energy is associated with the structure.")

        self._cache["total_energy_uncorrected"] = float(self.lines[line_ind].split()[5])
        return self._cache["total_energy_uncorrected"]

    @property
    def free_energy(self) -> float:
        """The total uncorrected energy from the aims.out file."""
        if "free_energy" in self._cache:
            return self._cache["free_energy"]
        
        try:
            line_ind = self.get_line_inds("free_energy")[-1]
        except IndexError:
            raise AimsParseError("No energy is associated with the structure.")

        self._cache["free_energy"] = float(self.lines[line_ind].split()[5])
        return self._cache["free_energy"]
    
    @property
    def total_energy(self) -> float:
        """The total energy of the structure"""
        if self.is_metallic and len(self.initial_geometry.lattice_vectors) == 3:
            return self.total_corrected_energy
        return self.total_uncorrected_energy

    @property
    def energy(self) -> float:
        """The force-consistent energy of the system"""
        return self.free_energy

    @property
    def dipole(self) -> Vector3D | None:
        """The electric dipole moment from the aims.out file."""
        if "dipole" in self._cache:
            return self._cache["dipole"]
        
        try:
            line_start = self.get_line_inds("dipole")[-1]
        except IndexError:
            self._cache["dipole"] = None
            return None

        print(self.lines[line_start])
        self._cache["dipole"] = np.array(
            [float(inp) for inp in self.lines[line_start].split()[6:9]]
        )
        return self._cache["dipole"]

    @property
    def dielectric_tensor(self) -> Matrix3D | None:
        """The dielectric tensor from the aims.out file."""
        if "dielectric_tensor" in self._cache:
            return self._cache["dielectric_tensor"]
        
        try:
            line_start = self.get_line_inds("dielectric_tensor")[-1]
        except IndexError:
            self._cache["dielectric_tensor"] = None
            return None

        # we should find the tensor in the next three lines:
        lines = self.lines[line_start + 1 : line_start + 4]

        # make ndarray and return
        self._cache["dielectric_tensor"] = np.array([np.fromstring(line, sep=" ") for line in lines])
        return self._cache["dielectric_tensor"]
    
    @property
    def polarization(self) -> Vector3D | None:
        """The polarization vector from the aims.out file."""
        if "polarization" in self._cache:
            return self._cache["polarization"]
        
        try:
            line_start = self.get_line_inds("polarization")[-1]
        except IndexError:
            self._cache["polarization"] = None
            return None
        
        self._cache["polarization"] = np.array([float(s) for s in self.lines[line_start].split()[-3:]])
        return self._cache["polarization"]


    def _parse_hirshfeld(
        self,
    ) -> None:
        """Parse the Hirshfled charges volumes, and dipole moments."""
        hirshfeld_charges = np.array(
            [float(self.lines[ind].split(":")[1]) for ind in self.get_line_inds("hirshfeld_charge")]
        )

        if len(hirshfeld_charges) != self.n_atoms:
            self._cache.update(
                {
                    "hirshfeld_charges": None,
                    "hirshfeld_volumes": None,
                    "hirshfeld_atomic_dipoles": None,
                    "hirshfeld_dipole": None,
                }
            )
            return
        
        hirshfeld_volumes = np.array(
            [float(self.lines[ind].split(":")[1]) for ind in self.get_line_inds("hirshfeld_volume")]
        )
        
        hirshfeld_atomic_dipoles = np.array(
            [[float(inp) for inp in self.lines[ind].split(":")[1].split()] for ind in self.get_line_inds("hirshfeld_dipole")]
        )

        if self.geometry.lattice_vectors is None:
            hirshfeld_dipole = np.sum(
                hirshfeld_charges.reshape((-1, 1)) * self.positions,
                axis=1,
            )
        else:
            hirshfeld_dipole = None

        self._cache.update(
            {
                "hirshfeld_charges": hirshfeld_charges,
                "hirshfeld_volumes": hirshfeld_volumes,
                "hirshfeld_atomic_dipoles": hirshfeld_atomic_dipoles,
                "hirshfeld_dipole": hirshfeld_dipole,
            }
        )

    def _parse_mulliken(
        self,
    ) -> None:
        """Parse the Mulliken charges and spins."""

        try:
            line_start = self.get_line_inds("mulliken_charges")[-1]
        except IndexError:
            self._cache.update(
                {
                    "mulliken_charges": None,
                    "mulliken_spins": None,
                }   
            )
            return
        
        mulliken_charges = np.array(
            [float(self.lines[ind].split()[3]) for ind in range(line_start + 3, line_start + 3 + self.n_atoms)]
        )

        try:
            line_start = self.get_line_inds("mulliken_spins")[-1]
        except IndexError:
            self._cache.update(
                {
                    "mulliken_charges": mulliken_charges,
                    "mulliken_spins": None,
                }   
            )
            return
        
        mulliken_spins = np.array(
            [float(self.lines[ind].split()[2]) for ind in range(line_start + 3, line_start + 3 + self.n_atoms)]
        )

        self._cache.update(
            {
                "mulliken_charges": mulliken_charges,
                "mulliken_spins": mulliken_spins,
            }
        )

    @property
    def geometry(self) -> AimsGeometry:
        """The pytmagen SiteCollection of the chunk."""
        if "geometry" not in self._cache:
            self._cache["geometry"] = self._parse_geometry()
        return self._cache["geometry"]

    @property
    def results(self) -> dict[str, Any]:
        """Get to a results dictionary for the section."""
        results = {
            "energy": self.energy,
            "free_energy": self.free_energy,
            "forces": self.forces,
            "stress": self.stress,
            "stresses": self.stresses,
            "magmom": self.magmom,
            "dipole": self.dipole,
            "fermi_energy": self.E_f,
            "n_iter": self.n_iter,
            "mulliken_charges": self.mulliken_charges,
            "mulliken_spins": self.mulliken_spins,
            "hirshfeld_charges": self.hirshfeld_charges,
            "hirshfeld_dipole": self.hirshfeld_dipole,
            "hirshfeld_volumes": self.hirshfeld_volumes,
            "hirshfeld_atomic_dipoles": self.hirshfeld_atomic_dipoles,
            "dielectric_tensor": self.dielectric_tensor,
            "polarization": self.polarization,
            "homo": self.homo,
            "lumo": self.lumo,
            "gap": self.gap,
            "direct_gap": self.direct_gap,
        }

        return {key: value for key, value in results.items() if value is not None}

    # Properties from the aims.out header
    @property
    def initial_geometry(self) -> AimsGeometry:
        """The initial structure for the calculation."""
        return self._header.initial_geometry

    @property
    def n_atoms(self) -> int:
        """The number of atoms in the structure."""
        return self._header.n_atoms

    @property
    def n_bands(self) -> int:
        """The number of Kohn-Sham states for the chunk."""
        return self._header.n_bands

    @property
    def n_electrons(self) -> int:
        """The number of electrons for the chunk."""
        return self._header.n_electrons

    @property
    def n_spins(self) -> int:
        """The number of spin channels for the chunk."""
        return self._header.n_spins

    @property
    def electronic_temperature(self) -> float:
        """The electronic temperature for the chunk."""
        return self._header.electronic_temperature

    @property
    def n_k_points(self) -> int:
        """The number of k_ppoints for the calculation."""
        return self._header.n_k_points

    @property
    def k_points(self) -> Sequence[Vector3D]:
        """All k-points listed in the calculation."""
        return self._header.k_points

    @property
    def k_point_weights(self) -> Sequence[float]:
        """The k-point weights for the calculation."""
        return self._header.k_point_weights

    @property
    def n_iter(self) -> int | None:
        """The number of steps needed to converge the SCF cycle for the chunk."""
        return self._parse_key_value_pair("number_of_iterations", allow_fail=True, dtype=int)

    @property
    def magmom(self) -> float | None:
        """The magnetic moment of the structure."""
        return self._parse_key_value_pair("magnetic_moment", allow_fail=True, dtype=float)

    @property
    def E_f(self) -> float | None:
        """The Fermi energy."""
        fermi_energy = self._parse_key_value_pair("fermi_energy", allow_fail=True)
        
        if isinstance(fermi_energy, str):
            self._cache["fermi_energy"] = float(fermi_energy.split(" eV")[0]) 

        return self._cache["fermi_energy"]

    @property
    def converged(self) -> bool:
        """True if the calculation is converged."""
        return (len(self.lines) > 0) and ("Have a nice day." in self.lines[-5:])

    @property
    def mulliken_charges(self) -> Sequence[float] | None:
        """The Mulliken charges of the system"""
        if "mulliken_charges" not in self._cache:
            self._parse_mulliken()
        return self._cache["mulliken_charges"]

    @property
    def mulliken_spins(self) -> Sequence[float] | None:
        """The Mulliken spins of the system"""
        if "mulliken_spins" not in self._cache:
            self._parse_mulliken()
        return self._cache["mulliken_spins"]

    @property
    def hirshfeld_charges(self) -> Sequence[float] | None:
        """The Hirshfeld charges of the system."""
        if "hirshfeld_charges" not in self._cache:
            self._parse_hirshfeld()
        return self._cache["hirshfeld_charges"]

    @property
    def hirshfeld_atomic_dipoles(self) -> Sequence[Vector3D] | None:
        """The Hirshfeld atomic dipoles of the system."""
        if "hirshfeld_atomic_dipoles" not in self._cache:
            self._parse_hirshfeld()
        return self._cache["hirshfeld_atomic_dipoles"]

    @property
    def hirshfeld_volumes(self) -> Sequence[float] | None:
        """The Hirshfeld atomic dipoles of the system."""
        if "hirshfeld_volumes" not in self._cache:
            self._parse_hirshfeld()
        return self._cache["hirshfeld_volumes"]

    @property
    def hirshfeld_dipole(self) -> None | Vector3D:
        """The Hirshfeld dipole of the system."""
        if "hirshfeld_dipole" not in self._cache:
            self._parse_hirshfeld()

        return self._cache["hirshfeld_dipole"]

    @property
    def homo(self) -> float:
        """The valance band maximum."""
        if "homo" in self._cache:
            return self._cache["homo"]
        
        if "homo" not in self._prop_line_relation:
            return None

        try:
            line_start = self.get_line_inds("homo")[-1]
        except IndexError:
            raise AimsParseError(f"No information about homo in the aims-output file")
        
        self._cache["homo"] = float(self.lines[line_start].split(" at ")[1].split(" eV")[0].strip())
        return self._cache["homo"]

    @property
    def lumo(self) -> float:
        """The conduction band minimnum."""
        if "lumo" in self._cache:
            return self._cache["lumo"]
        
        if "lumo" not in self._prop_line_relation:
            return None

        try:
            line_start = self.get_line_inds("lumo")[-1]
        except IndexError:
            raise AimsParseError(f"No information about lumo in the aims-output file")
        
        self._cache["lumo"] = float(self.lines[line_start].split(" at ")[1].split(" eV")[0].strip())
        return self._cache["lumo"]

    @property
    def gap(self) -> float:
        """The band gap."""
        return float(self._parse_key_value_pair("gap").split(" eV")[0])

    @property
    def direct_gap(self) -> float:
        """The direct bandgap."""
        direct_gap = self._parse_key_value_pair("direct_gap", allow_fail=True)
        if direct_gap is None:
            return self.gap
        
        return float(direct_gap.split(" eV")[0])




from __future__ import annotations

import gzip
from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_allclose
from pyfhiaims.output_parser.aims_out_section import (
    AimsParseError,
    AimsOutSection
)
from pyfhiaims.output_parser.aims_out_header_section import AimsOutHeaderSection
from pyfhiaims.output_parser.aims_out_calculation_section import AimsOutCalcSection

eps_hp = 1e-15  # The epsilon value used to compare numbers that are high-precision
eps_lp = 1e-7  # The epsilon value used to compare numbers that are low-precision

parser_file_dir = Path(__file__).parent / "parser_checks"


@pytest.fixture()
def default_section():
    lines = ["TEST", "A", "TEST", "| Number of atoms: 200"]
    return AimsOutSection(lines)


def test_get_line_inds(default_section):
    test_keys = {"test": ["TEST"], "a": "A", "n_atoms": "| Number of atoms", "test_A": "TEST A"}

    default_section.set_prop_line_relations(test_keys)

    assert default_section.get_line_inds("test") == [0, 2]
    assert default_section.get_line_inds("test_A") == []
    assert default_section.get_line_inds("a") == [1]


def test_parse_key_value_pair(default_section):
    test_keys = {"test": ["TEST"], "a": "A", "n_atoms": "| Number of atoms", "n_electrons": "Number of electrons"}
    default_section.set_prop_line_relations(test_keys)
    assert default_section._parse_key_value_pair("n_atoms", dtype=int) == 200
    assert default_section._parse_key_value_pair("n_bands", dtype=int) is None
    assert default_section._parse_key_value_pair("n_electrons", allow_fail=True, dtype=int) is None


@pytest.fixture()
def empty_header_section():
    return AimsOutHeaderSection([])


@pytest.mark.parametrize("attr_name", ["n_atoms", "n_bands", "n_electrons", "n_spins", "initial_geometry", "initial_lattice", "initial_charges", "initial_magnetic_moments"])
def test_missing_parameter(attr_name, empty_header_section):
    with pytest.raises(AimsParseError, match="No information about"):
        getattr(empty_header_section, attr_name)


def test_default_header_electronic_temperature(empty_header_section):
    assert empty_header_section.electronic_temperature == 0.0


def test_default_header_is_md(empty_header_section):
    assert not empty_header_section.is_md


def test_default_header_is_relaxation(empty_header_section):
    assert not empty_header_section.is_relaxation


def test_default_header_n_k_points(empty_header_section):
    assert empty_header_section.n_k_points is None


def test_default_header_k_points(empty_header_section):
    assert empty_header_section.k_points is None


def test_default_header_k_point_weights(empty_header_section):
    assert empty_header_section.k_point_weights is None


@pytest.fixture()
def initial_lattice():
    return np.array(
        [
            [1, 2.70300000, 3.70300000],
            [4.70300000, 2, 6.70300000],
            [8.70300000, 7.70300000, 3],
        ]
    )


@pytest.fixture()
def header_section():
    with open(f"{parser_file_dir}/header_section.out", mode="rt") as hc_file:
        lines = hc_file.readlines()

    for ll, line in enumerate(lines):
        lines[ll] = line.strip()

    return AimsOutHeaderSection(lines)


def test_header_n_atoms(header_section):
    assert header_section.n_atoms == 2


def test_header_n_bands(header_section):
    assert header_section.n_bands == 3


def test_header_n_electrons(header_section):
    assert header_section.n_electrons == 28


def test_header_n_spins(header_section):
    assert header_section.n_spins == 2


def test_header_initial_geometry(header_section, initial_lattice):
    initial_positions = np.array([[0.000, 1.000, 2.000], [2.703, 3.703, 4.703]])
    assert len(header_section.initial_geometry) == 2
    assert_allclose(header_section.initial_geometry.lattice_vectors, initial_lattice)
    assert_allclose(header_section.initial_geometry.positions, initial_positions, atol=1e-12)
    assert [sym for sym in header_section.initial_geometry.symbols] == ["Na", "Cl"]


def test_header_electronic_temperature(header_section):
    assert header_section.electronic_temperature == 0.05


def test_header_is_md(header_section):
    assert header_section.is_md


def test_header_is_relaxation(header_section):
    assert header_section.is_relaxation


def test_header_n_k_points(header_section):
    assert header_section.n_k_points == 8


@pytest.fixture()
def k_points():
    return np.array(
        [
            [0.000, 0.000, 0.000],
            [0.000, 0.000, 0.500],
            [0.000, 0.500, 0.000],
            [0.000, 0.500, 0.500],
            [0.500, 0.000, 0.000],
            [0.500, 0.000, 0.500],
            [0.500, 0.500, 0.000],
            [0.500, 0.500, 0.500],
        ]
    )


def test_header_k_point_weights(
    header_section,
):
    assert_allclose(header_section.k_point_weights, np.full((8), 0.125))


def test_header_k_points(header_section, k_points):
    assert_allclose(header_section.k_points, k_points)


def test_header_header_summary(header_section, k_points):
    header_summary = {
        "initial_geometry": header_section.initial_geometry,
        "initial_lattice": header_section.initial_lattice,
        "is_relaxation": True,
        "is_md": True,
        "n_atoms": 2,
        "n_bands": 3,
        "n_electrons": 28,
        "n_spins": 2,
        "electronic_temperature": 0.05,
        "n_k_points": 8,
        "k_points": k_points,
        "k_point_weights": np.full((8), 0.125),
    }
    for key, val in header_section.header_summary.items():
        if isinstance(val, np.ndarray):
            assert_allclose(val, header_summary[key])
        else:
            assert val == header_summary[key]


@pytest.fixture()
def empty_calc_section(header_section):
    return AimsOutCalcSection([], header_section)


def test_header_transfer_n_atoms(empty_calc_section):
    assert empty_calc_section.n_atoms == 2


def test_header_transfer_n_bands(empty_calc_section):
    assert empty_calc_section.n_bands == 3


def test_header_transfer_n_electrons(empty_calc_section):
    assert empty_calc_section.n_electrons == 28


def test_header_transfer_n_spins(empty_calc_section):
    assert empty_calc_section.n_spins == 2


def test_header_transfer_initial_geometry(empty_calc_section, initial_lattice):
    initial_positions = np.array([[0.000, 1.000, 2.000], [2.703, 3.703, 4.703]])

    assert len(empty_calc_section.initial_geometry) == 2
    assert_allclose(empty_calc_section.initial_geometry.lattice_vectors, initial_lattice)
    assert_allclose(empty_calc_section.initial_geometry.positions, initial_positions, atol=1e-12)
    assert [sym for sym in empty_calc_section.initial_geometry.symbols] == ["Na", "Cl"]


def test_header_transfer_electronic_temperature(empty_calc_section):
    assert empty_calc_section.electronic_temperature == 0.05


def test_header_transfer_n_k_points(empty_calc_section):
    assert empty_calc_section.n_k_points == 8


def test_header_transfer_k_point_weights(empty_calc_section):
    assert_allclose(empty_calc_section.k_point_weights, np.full((8), 0.125))


def test_header_transfer_k_points(empty_calc_section, k_points):
    assert_allclose(empty_calc_section.k_points, k_points)


def test_default_calc_energy_raises_error(empty_calc_section):
    with pytest.raises(AimsParseError, match="No energy is associated with the structure."):
        _ = empty_calc_section.energy

def test_default_calc_total_energy_raises_error(empty_calc_section):
    with pytest.raises(AimsParseError, match="No energy is associated with the structure."):
        _ = empty_calc_section.total_energy

@pytest.mark.parametrize(
    "attr",
    [
        "forces",
        "stresses",
        "stress",
        "n_iter",
        "magmom",
        "E_f",
        "dipole",
        "mulliken_charges",
        "mulliken_spins",
        "hirshfeld_charges",
        "hirshfeld_volumes",
        "hirshfeld_atomic_dipoles",
        "hirshfeld_dipole",
    ],
)
def test_section_defaults_none(attr, empty_calc_section):
    assert getattr(empty_calc_section, attr) is None


def test_default_calc_is_metallic(empty_calc_section):
    assert not empty_calc_section.is_metallic


def test_default_calc_converged(empty_calc_section):
    assert not empty_calc_section.converged


@pytest.fixture()
def calc_section(header_section):
    with open(f"{parser_file_dir}/calc_section.out", mode="rt") as file:
        lines = file.readlines()

    for ll, line in enumerate(lines):
        lines[ll] = line.strip()
    return AimsOutCalcSection(lines, header_section)


@pytest.fixture()
def numerical_stress_section(header_section):
    with open(f"{parser_file_dir}/numerical_stress.out", mode="rt") as file:
        lines = file.readlines()

    for ll, line in enumerate(lines):
        lines[ll] = line.strip()
    return AimsOutCalcSection(lines, header_section)


def test_calc_structure(calc_section, initial_lattice):
    initial_positions = np.array([[0.000, 1.000, 2.000], [2.703, 3.703, 4.703]])

    assert len(calc_section.geometry) == 2
    assert_allclose(calc_section.geometry.lattice_vectors, initial_lattice)
    assert_allclose(calc_section.geometry.positions, initial_positions, atol=1e-12)
    assert [sym for sym in calc_section.geometry.symbols] == ["Na", "Cl"]


def test_calc_forces(calc_section):
    forces = np.array([[1.0, 2.0, 3.0], [6.0, 5.0, 4.0]])
    assert_allclose(calc_section.forces, forces)

    # Different because of the constraints
    assert_allclose(calc_section.results["forces"], forces)


# def test_calc_stresses(calc_section):
#     stresses = 1.0 * np.array(
#         [
#             Tensor.from_voigt([-10.0, -20.0, -30.0, -60.0, -50.0, -40.0]),
#             Tensor.from_voigt([10.0, 20.0, 30.0, 60.0, 50.0, 40.0]),
#         ]
#     )
#     assert_allclose(calc_section.stresses, stresses)
#     assert_allclose(calc_section.results["stresses"], stresses)


def test_calc_stress(calc_section):
    stress = 1.0 * np.array([[1.0, 2.0, 3.0], [2.0, 5.0, 6.0], [3.0, 6.0, 7.0]])
    assert_allclose(calc_section.stress, stress)
    assert_allclose(calc_section.results["stress"], stress)


def test_calc_num_stress(numerical_stress_section):
    stress = 1.0 * np.array([[1.0, 2.0, 3.0], [2.0, 5.0, 6.0], [3.0, 6.0, 7.0]])
    assert_allclose(numerical_stress_section.stress, stress)
    assert_allclose(numerical_stress_section.results["stress"], stress)


def test_calc_free_energy(calc_section):
    free_energy = -3.169503986610555e05
    assert np.abs(calc_section.free_energy - free_energy) < eps_hp
    assert np.abs(calc_section.results["free_energy"] - free_energy) < eps_hp


def test_calc_energy(calc_section):
    energy = -3.169503986610555e05
    assert np.abs(calc_section.energy - energy) < eps_hp
    assert np.abs(calc_section.results["energy"] - energy) < eps_hp


def test_calc_magnetic_moment(calc_section):
    magmom = 0
    assert calc_section.magmom == magmom
    assert calc_section.results["magmom"] == magmom


def test_calc_n_iter(calc_section):
    n_iter = 58
    assert calc_section.n_iter == n_iter
    assert calc_section.results["n_iter"] == n_iter


def test_calc_fermi_energy(calc_section):
    Ef = -8.24271207
    assert np.abs(calc_section.E_f - Ef) < eps_lp
    assert np.abs(calc_section.results["fermi_energy"] - Ef) < eps_lp


def test_calc_dipole(calc_section):
    assert calc_section.dipole is None


def test_calc_is_metallic(calc_section):
    assert calc_section.is_metallic


def test_calc_converged(calc_section):
    assert calc_section.converged


def test_calc_mulliken_charges(calc_section):
    mulliken_charges = [0.617623, -0.617623]
    assert_allclose(calc_section.mulliken_charges, mulliken_charges)
    assert_allclose(calc_section.results["mulliken_charges"], mulliken_charges)


def test_calc_mulliken_spins(calc_section):
    # TARP: False numbers added to test parsing
    mulliken_spins = [-0.003141, 0.002718]
    assert_allclose(calc_section.mulliken_spins, mulliken_spins)
    assert_allclose(calc_section.results["mulliken_spins"], mulliken_spins)


def test_calc_hirshfeld_charges(calc_section):
    hirshfeld_charges = [0.20898543, -0.20840994]
    assert_allclose(calc_section.hirshfeld_charges, hirshfeld_charges)
    assert_allclose(calc_section.results["hirshfeld_charges"], hirshfeld_charges)


def test_calc_hirshfeld_volumes(calc_section):
    hirshfeld_volumes = [73.39467444, 62.86011074]
    assert_allclose(calc_section.hirshfeld_volumes, hirshfeld_volumes)
    assert_allclose(calc_section.results["hirshfeld_volumes"], hirshfeld_volumes)


def test_calc_hirshfeld_atomic_dipoles(calc_section):
    hirshfeld_atomic_dipoles = np.zeros((2, 3))
    assert_allclose(calc_section.hirshfeld_atomic_dipoles, hirshfeld_atomic_dipoles)
    assert_allclose(calc_section.results["hirshfeld_atomic_dipoles"], hirshfeld_atomic_dipoles)


def test_calc_hirshfeld_dipole(calc_section):
    assert calc_section.hirshfeld_dipole is None


@pytest.fixture()
def molecular_header_section():
    with open(f"{parser_file_dir}/molecular_header_section.out", mode="rt") as file:
        lines = file.readlines()

    for ll, line in enumerate(lines):
        lines[ll] = line.strip()

    return AimsOutHeaderSection(lines)


@pytest.mark.parametrize(
    "attr_name",
    ["k_points", "k_point_weights", "n_k_points"],
)
def test_section_molecular_header_defaults_none(attr_name, molecular_header_section):
    assert getattr(molecular_header_section, attr_name) is None


def test_molecular_header_n_bands(molecular_header_section):
    assert molecular_header_section.n_bands == 7


def test_molecular_header_initial_geometry(molecular_header_section, molecular_positions):
    assert len(molecular_header_section.initial_geometry) == 3
    assert [sym for sym in molecular_header_section.initial_geometry.symbols] == ["O", "H", "H"]
    assert_allclose(
        molecular_header_section.initial_geometry.positions,
        [[0, 0, 0], [0.95840000, 0, 0], [-0.24000000, 0.92790000, 0]],
    )


@pytest.fixture()
def molecular_calc_section(molecular_header_section):
    with open(f"{parser_file_dir}/molecular_calc_section.out", mode="rt") as file:
        lines = file.readlines()

    for idx, line in enumerate(lines):
        lines[idx] = line.strip()
    return AimsOutCalcSection(lines, molecular_header_section)


@pytest.fixture()
def molecular_positions():
    return np.array([[-0.00191785, -0.00243279, 0], [0.97071531, -0.00756333, 0], [-0.25039746, 0.93789612, 0]])


def test_molecular_calc_atoms(molecular_calc_section, molecular_positions):
    print(molecular_calc_section.geometry)
    assert len(molecular_calc_section.geometry.symbols) == 3
    assert_allclose(molecular_calc_section.geometry.positions, molecular_positions)
    assert [sym for sym in molecular_calc_section.geometry.symbols] == ["O", "H", "H"]


def test_molecular_calc_forces(molecular_calc_section):
    forces = np.array(
        [
            [0.502371357164392e-03, 0.518627676606471e-03, 0.000000000000000e00],
            [-0.108826758257187e-03, -0.408128912334209e-03, -0.649037698626122e-27],
            [-0.393544598907207e-03, -0.110498764272267e-03, -0.973556547939183e-27],
        ]
    )
    assert_allclose(molecular_calc_section.forces, forces)
    assert_allclose(molecular_calc_section.results["forces"], forces)


@pytest.mark.parametrize("attrname", ["stresses", "stress", "magmom", "E_f"])
def test_section_molecular_defaults_none(attrname, molecular_calc_section):
    assert getattr(molecular_calc_section, attrname) is None


def test_molecular_calc_free_energy(molecular_calc_section):
    free_energy = -2.206778551123339e04
    assert np.abs(molecular_calc_section.free_energy - free_energy) < eps_hp
    assert np.abs(molecular_calc_section.results["free_energy"] - free_energy) < eps_hp


def test_molecular_calc_energy(molecular_calc_section):
    energy = -2.206778551123339e04
    assert np.abs(molecular_calc_section.energy - energy) < eps_hp
    assert np.abs(molecular_calc_section.results["energy"] - energy) < eps_hp


def test_molecular_calc_n_iter(molecular_calc_section):
    n_iter = 7
    assert molecular_calc_section.n_iter == n_iter
    assert molecular_calc_section.results["n_iter"] == n_iter


def test_molecular_calc_dipole(molecular_calc_section):
    dipole = [0.260286493869765, 0.336152447755231, 0.470003778119121e-15]
    assert_allclose(molecular_calc_section.dipole, dipole)
    assert_allclose(molecular_calc_section.results["dipole"], dipole)


def test_molecular_calc_is_metallic(molecular_calc_section):
    assert not molecular_calc_section.is_metallic


def test_molecular_calc_converged(molecular_calc_section):
    assert molecular_calc_section.converged


def test_molecular_calc_hirshfeld_charges(molecular_calc_section):
    molecular_hirshfeld_charges = np.array([-0.32053200, 0.16022630, 0.16020375])
    assert_allclose(molecular_calc_section.hirshfeld_charges, molecular_hirshfeld_charges)
    assert_allclose(molecular_calc_section.results["hirshfeld_charges"], molecular_hirshfeld_charges)


def test_molecular_calc_hirshfeld_volumes(molecular_calc_section):
    hirshfeld_volumes = np.array([21.83060659, 6.07674041, 6.07684447])
    assert_allclose(molecular_calc_section.hirshfeld_volumes, hirshfeld_volumes)
    assert_allclose(molecular_calc_section.results["hirshfeld_volumes"], hirshfeld_volumes)


def test_molecular_calc_hirshfeld_atomic_dipoles(molecular_calc_section):
    hirshfeld_atomic_dipoles = np.array(
        [[0.04249319, 0.05486053, 0], [0.13710134, -0.00105126, 0], [-0.03534982, 0.13248706, 0]]
    )
    assert_allclose(molecular_calc_section.hirshfeld_atomic_dipoles, hirshfeld_atomic_dipoles)
    assert_allclose(
        molecular_calc_section.results["hirshfeld_atomic_dipoles"],
        hirshfeld_atomic_dipoles,
    )


def test_molecular_calc_hirshfeld_dipole(molecular_calc_section, molecular_positions):
    molecular_hirshfeld_charges = np.array([-0.32053200, 0.16022630, 0.16020375])
    hirshfeld_dipole = np.sum(molecular_hirshfeld_charges.reshape((-1, 1)) * molecular_positions, axis=1)

    assert_allclose(molecular_calc_section.hirshfeld_dipole, hirshfeld_dipole)
    assert_allclose(molecular_calc_section.results["hirshfeld_dipole"], hirshfeld_dipole)



"""FHI-aims objects representing K-points (in the form of grid or list)."""

from dataclasses import dataclass
from enum import Enum


class KPointsType(Enum):
    GRID = 'k_grid'
    DENSITY = 'k_grid_density'
    EXTERNAL = 'k_points_external'


@dataclass
class AimsKPoints:
    """A base class for FHI-aims K-points representation."""
    type: KPointsType



"""Type casting utils from pymatgen that are useful"""
from typing import TypeVar, Sequence, Literal, Annotated
import numpy as np
import numpy.typing as npt


_T = TypeVar("_T", bound=np.generic)

Vector3D = Annotated[npt.NDArray[_T], Literal[3]]
Matrix3D = Annotated[npt.NDArray[_T], Literal[3, 3]]


def to_vector3d(sequence: Sequence, dtype=float) -> Vector3D:
    """Type casting to 3D vector of floats."""
    assert len(sequence) == 3
    return np.array(sequence, dtype=dtype)

def to_matrix3d(sequence: Sequence, dtype=float) -> Matrix3D:
    """Type casting to 3D matrix of floats."""
    assert len(sequence) == 3 and all([len(row) == 3 for row in sequence])
    return np.array(sequence, dtype=dtype)



"""Tests for AimsControlIn object."""

from pathlib import Path

from pyfhiaims.control import AimsControlIn


def test_control(data_dir):
    control_dir = data_dir / 'control'
    control_in = AimsControlIn.from_file(control_dir / "al2zns4.in")
    assert control_in["output_level"] == "MD_light"
    assert control_in["xc"] == "pw-lda"
    assert "Al" in control_in.species_defaults
    # print(control_in.species_defaults["Co"].content)


"""Various utilities for pyaims"""


def is_number_string(s: str) -> bool:
    """Checks if s is a string made of float numbers and optionally comment sign.

    Parameters:
        s: str
            A string to check

    """
    allowed_chars = set("-.# ")
    return all(char.isdigit() or char in allowed_chars for char in s)


"""AIMS output parser for the header section, taken from ASE with modifications."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
import numpy as np

from pyfhiaims.geometry.geometry import AimsGeometry
from pyfhiaims.geometry.geometry import Matrix3D
from pyfhiaims.output_parser.aims_out_section import AimsOutSection, AimsParseError
if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from pymatgen.util.typing import Vector3D

__author__ = "Thomas A. R. Purcell and Andrey Sobolev"
__version__ = "1.0"
__email__ = "purcellt@arizona.edu and andrey.n.sobolev@gmail.com"
__date__ = "July 2024"

# TARP: Originally an object, but type hinting needs this to be an int
LINE_NOT_FOUND = -1000
EV_PER_A3_TO_KBAR = 1.60217653e-19 * 1e22

PARSE_HEADER_KEYS: dict[str, list[str]] = {
    "commit_hash": ["Commit number"],
    "aims_uuid": ["aims_uuid"],
    "version_number": ["FHI-aims version"],
    "fortran_compiler": ["Fortran compiler      :"],
    "c_compiler": ["C compiler            :"],
    "fortran_compiler_flags": ["Fortran compiler flags"],
    "c_compiler_flags": ["C compiler flags"],
    "linked_against": ["Linking against:"],
    "initial_geometry": ["Parsing geometry.in (first pass over file, find array dimensions only)."],
    "is_md": ["Complete information for previous time-step:"],
    "is_relaxation": ["Geometry relaxation:"],
    "n_atoms": ["| Number of atoms"],
    "n_bands": [
        "Number of Kohn-Sham states",
        "Reducing total number of  Kohn-Sham states",
        "Reducing total number of Kohn-Sham states",
    ],
    "n_electrons": ["The structure contains"],
    "n_kpts": ["| Number of k-points"],
    "n_spins": ["| Number of spin channels"],
    "electronic_temp": ["Occupation type:"],
    "k_points": ["| K-points in task"],
    "end_k_points": ["| k-point:"],
    "build_type": ["Using"],
}

@dataclass
class AimsOutHeaderSection(AimsOutSection):
    """The header of the aims.out file containing general information."""

    def __post_init__(self):
        """Setup line parsing"""
        self.set_prop_line_relations(PARSE_HEADER_KEYS)
    
    @property
    def commit_hash(self) -> str:
        """The commit hash for the FHI-aims version."""
        return self._parse_key_value_pair("commit_hash")

    @property
    def aims_uuid(self) -> str:
        """The aims-uuid for the calculation."""
        return self._parse_key_value_pair("aims_uuid")

    @property
    def version_number(self) -> str:
        """The version number for the FHI-aims version."""
        return self._parse_key_value_pair("version_number")

    @property
    def fortran_compiler(self) -> str | None:
        """The fortran compiler used to make FHI-aims."""
        return self._parse_key_value_pair("fortran_compiler")

    @property
    def c_compiler(self) -> str | None:
        """The C compiler used to make FHI-aims."""
        return self._parse_key_value_pair("c_compiler", True)

    @property
    def fortran_compiler_flags(self) -> str | None:
        """The fortran compiler flags used to make FHI-aims."""
        return self._parse_key_value_pair("fortran_compiler_flags")

    @property
    def c_compiler_flags(self) -> str | None:
        """The C compiler flags used to make FHI-aims."""
        return self._parse_key_value_pair("c_compiler_flags", True)

    @property
    def build_type(self) -> list[str]:
        """The optional build flags passed to cmake."""
        if "build_type" in self._cache:
            return self._cache["build_type"]
        
        line_inds = self.get_line_inds("build_type")
        end_line = 0
        while end_line < len(line_inds) and "parallel tasks." not in self.lines[line_inds[end_line]]:
            end_line += 1

        self._cache["build_type"] = [" ".join(self.lines[line_inds[ind]].split()[1:]).strip() for ind in line_inds[:end_line]]
        return self._cache["build_type"]

    @property
    def linked_against(self) -> list[str]:
        """All libraries used to link the FHI-aims executable."""
        if "linked_against" in self._cache:
            return self._cache["linked_against"]
        
        try:
            line_start = self.get_line_inds("linked_against")[-1]
        except IndexError:
            self._cache["linked_against"] = []
            return []

        linked_libs = [self.lines[line_start].split(":")[1].strip()]
        line_start += 1
        while "lib" in self.lines[line_start]:
            linked_libs.append(self.lines[line_start].strip())
            line_start += 1

        self._cache["linked_against"] = linked_libs
        return linked_libs

    @property
    def initial_lattice(self) -> Matrix3D | None:
        """The initial lattice vectors from the aims.out file."""
        return self.initial_geometry.lattice_vectors

    @property
    def initial_geometry(self) -> AimsGeometry:
        """The initial structure.

        Using the FHI-aims output file recreate the initial structure for
        the calculation.
        """
        if "initial_geometry" in self._cache:
            return self._cache["initial_geometry"]
        
        try:
            line_start = self.get_line_inds("initial_geometry")[-1]
        except IndexError:
            raise AimsParseError("No information about the geometry in the section.")

        while "---------" not in self.lines[line_start]:
            line_start += 1
        
        line_start += 1
        line_end = line_start + 1
        while "---------" not in self.lines[line_end]:
            line_end += 1
        
        self._cache["initial_geometry"] = AimsGeometry.from_strings(self.lines[line_start:line_end])
        return self._cache["initial_geometry"]

    @property
    def initial_charges(self) -> Sequence[float]:
        """The initial charges for the structure."""
        return self.initial_geometry.initial_charge

    @property
    def initial_magnetic_moments(self) -> Sequence[float]:
        """The initial magnetic Moments."""
        return self.initial_geometry.initial_moment

    @property
    def is_md(self) -> bool:
        """Is the output for a molecular dynamics calculation?"""
        return len(self.get_line_inds("is_md")) > 0

    @property
    def is_relaxation(self) -> bool:
        """Is the output for a relaxation?"""
        return len(self.get_line_inds("is_relaxation")) > 0

    def _parse_k_points(self) -> None:
        """Parse the list of k-points used in the calculation."""
        n_kpts = self.n_k_points
        if n_kpts is None:
            self._cache.update(
                {
                    "k_points": None,
                    "k_point_weights": None,
                }
            )
            return

        try:
            line_start = self.get_line_inds("k_points")[-1]
            line_end = self.get_line_inds("end_k_points")[-1]
        except IndexError:
            self._cache.update(
                {
                    "k_points": None,
                    "k_point_weights": None,
                }
            )
            return

        k_points = np.zeros((n_kpts, 3))
        k_point_weights = np.zeros(n_kpts)
        for kk, line in enumerate(self.lines[line_start + 1 : line_end + 1]):
            k_points[kk] = [float(inp) for inp in line.split()[4:7]]
            k_point_weights[kk] = float(line.split()[-1])

        self._cache.update(
            {
                "k_points": k_points,
                "k_point_weights": k_point_weights,
            }
        )

    @property
    def n_atoms(self) -> int:
        """The number of atoms for the material."""
        return self._parse_key_value_pair("n_atoms", dtype=int)


    @property
    def n_bands(self) -> int | None:
        """The number of Kohn-Sham states for the chunk."""
        if "n_bands" in self._cache:
            return self._cache["n_bands"]
        
        if "n_bands" not in self._prop_line_relation:
            return None

        try:
            line_start = self.get_line_inds("n_bands")[-1]
        except IndexError:
            raise AimsParseError(f"No information about n_bands in the aims-output file")
        
        self._cache["n_bands"] = int(self.lines[line_start].split(" ")[-1].split(".")[0].strip())
        return self._cache["n_bands"]

    @property
    def n_electrons(self) -> int | None:
        """The number of electrons for the chunk."""
        if "n_electrons" in self._cache:
            return self._cache["n_electrons"]
        
        try:
            line_start = self.get_line_inds("n_electrons")[-1]
        except IndexError:
            raise AimsParseError(f"No information about n_electrons in the aims-output file")
        
        self._cache["n_electrons"] = float(self.lines[line_start].split()[-2].strip())
        return self._cache["n_electrons"]

    @property
    def n_k_points(self) -> int | None:
        """The number of k_ppoints for the calculation."""
        n_kpts = self._parse_key_value_pair("n_kpts", allow_fail=True, dtype=int)
        if isinstance(n_kpts, str):
            self._cache["n_kpts"] = int(self._cache["n_kpts"])
        return self._cache["n_kpts"]

    @property
    def n_spins(self) -> int | None:
        """The number of spin channels for the chunk."""
        return self._parse_key_value_pair("n_spins", dtype=int)

    @property
    def electronic_temperature(self) -> float:
        """The electronic temperature for the chunk."""
        if "electronic_temperature" in self._cache:
            return self._cache["electronic_temperature"]
        
        try:
            line_start = self.get_line_inds("electronic_temp")[-1]
        except IndexError:
            # TARP: Default FHI-aims value
            self._cache["electronic_temperature"] = 0.0
            return 0.00

        line = self.lines[line_start]
        self._cache["electronic_temperature"] =  float(line.split("=")[-1].strip().split()[0])
        return self._cache["electronic_temperature"]

    @property
    def k_points(self) -> Sequence[Vector3D]:
        """All k-points listed in the calculation."""
        if "k_points" not in self._cache:
            self._parse_k_points()

        return self._cache["k_points"]

    @property
    def k_point_weights(self) -> Sequence[float]:
        """The k-point weights for the calculation."""
        if "k_point_weights" not in self._cache:
            self._parse_k_points()

        return self._cache["k_point_weights"]

    @property
    def header_summary(self) -> dict[str, Any]:
        """Dictionary summarizing the information inside the header."""
        return {
            "initial_geometry": self.initial_geometry,
            "initial_lattice": self.initial_lattice,
            "is_relaxation": self.is_relaxation,
            "is_md": self.is_md,
            "n_atoms": self.n_atoms,
            "n_bands": self.n_bands,
            "n_electrons": self.n_electrons,
            "n_spins": self.n_spins,
            "electronic_temperature": self.electronic_temperature,
            "n_k_points": self.n_k_points,
            "k_points": self.k_points,
            "k_point_weights": self.k_point_weights,
        }

    @property
    def metadata_summary(self) -> dict[str, list[str] | str | None]:
        """Dictionary containing all metadata for FHI-aims build."""
        return {
            "commit_hash": self.commit_hash,
            "aims_uuid": self.aims_uuid,
            "version_number": self.version_number,
            "fortran_compiler": self.fortran_compiler,
            "c_compiler": self.c_compiler,
            "fortran_compiler_flags": self.fortran_compiler_flags,
            "c_compiler_flags": self.c_compiler_flags,
            "build_type": self.build_type,
            "linked_against": self.linked_against,
        }




"""Some tests for the AimsStdout object."""

from math import isclose

from pyfhiaims import AimsStdout


def test_stdout(data_dir):
    stdout_data_dir = data_dir / "stdout"
    file_name = stdout_data_dir / "relax.out.gz"
    stdout = AimsStdout(file_name)
    assert len(stdout.warnings) == 1
    assert len(stdout.errors) == 0
    assert stdout.is_finished_ok
    assert stdout.final["geometry_converged"]
    assert isclose(stdout.energy, -1071304.321815)
    assert stdout.forces is not None
    assert stdout.metadata["num_atoms"] == 2







""" A main output file parser for FHI-aims. Realizes the Deterministic State Machine pattern
"""
from __future__ import annotations

import importlib
import random
import string
from abc import abstractmethod
import inspect
import re
from pathlib import Path
from typing import Any, IO, Type

from monty.io import zopen

from .converters import to_str
from .utils import update


def parser_classes(module) -> dict[str, Type[ChunkParser]]:
    """Returns the dict of all the file and chunk parser classes declared in
    the current module"""
    result = {}
    for _, cls in inspect.getmembers(module, inspect.isclass):
        if issubclass(cls, ChunkParser):
            result[cls.name] = cls
    return result


def _get_parser(parser_class, header_line):
    """A factory that injects the header line into the ChunkParser instance"""
    def _factory(file_descriptor: str | Path | IO,
                 parent: Parser = None) -> Parser:
        return parser_class(file_descriptor, parent, header_line)
    return _factory


class Parser:

    name: str = ""

    def __init__(self,
                 file_descriptor: str | Path | IO,
                 parent: FileParser = None) -> None:
        """ A base class for the parser. Gets the

        Parameters
        ----------
        file_descriptor: str | Path | IO
            A name or handle of the main output file
        parent: Parser
            A parent parser for the current instance
        """
        # get the module in which the actual implementation is located
        # (parsers.py) and available parsers within that module
        module = importlib.import_module(self.__module__)
        self._available_parsers = parser_classes(module)
        self.result = {}
        self.warnings = []
        self.errors = []
        # propagate metadata along inheritance line
        self.parent = parent
        if self.parent is None:
            self.run_metadata = {}
        else:
            self.run_metadata = self.parent.run_metadata
        self.next = None
        # open the file to parse
        if isinstance(file_descriptor, (str, Path)):
            self.fd = zopen(file_descriptor, "rt")
            self.is_root = True
        else:
            self.fd = file_descriptor
            self.is_root = False

    def __del__(self):
        """Close the file if it is the root parser"""
        if self.is_root:
            self.fd.close()

    @abstractmethod
    def parse(self):
        raise NotImplementedError

    @abstractmethod
    def annotate(self, annotated_file: str | Path | IO):
         raise NotImplementedError


class FileParser(Parser):

    initial_chunk: str

    def __init__(self,
                 file_descriptor: str | Path | IO,
                 parent: Parser = None,
                 header_line: str = None) -> None:
        """
        A main Parser object.

        Parameters
        ----------
        file_descriptor : str | Path | IO
            A name or handle of the main output file
        """
        super().__init__(file_descriptor, parent)
        # inject header line into the ChunkParser class
        self._initial_chunk = _get_parser(self._available_parsers[self.initial_chunk],
                                          header_line)

    def parse(self) -> dict:
        """ Parses the output file by dividing it into chunks, and then parsing them separately

        Returns
        -------
            a parsed dictionary
        """
        self.fd.seek(0)
        # instantiate ChunkParser with header line already injected
        current_chunk = self._initial_chunk(self.fd, parent=self)
        # can we create the `parse_results` structure preliminary?
        parse_results = {}
        while current_chunk:
            chunk_results = current_chunk.parse()
            if chunk_results or current_chunk.parsed_key:
                update(parse_results, chunk_results, current_chunk.parsed_key)
            if current_chunk.next is None:
                break
            current_chunk = current_chunk.next(self.fd, parent=self)
        self.result = parse_results
        return self.result

    def annotate(self, annotated_file: str | Path | IO = None) -> None:
        """Annotates the output file by adding the chunk name and ID to each line.

        Parameters
        ----------
        annotated_file: str | Path | IO
            The name of the file to write annotations to.
            If None, write the annotated file to stdout.
        """
        self.fd.seek(0)
        if isinstance(annotated_file, str | Path):
            annotated_fd = open(annotated_file, "w")
        else:
            annotated_fd = annotated_file
        current_chunk = self._initial_chunk(self.fd, parent=self)
        while current_chunk:
            current_chunk.annotate(annotated_fd)
            if current_chunk.next is None:
                break
            current_chunk = current_chunk.next(self.fd, parent=self)
        if isinstance(annotated_fd, IO):
            annotated_fd.close()


class ChunkParser(Parser):
    """ An abstract class for the output file chunks

    Attributes
    ----------
    name : str
        the name of the section in the parse results
    values : dict[str, Any]
        the dict of the names and regular expressions that may be found in the
        chunk
    next_chunks : list[str | dict]
        the collection of lines and runtime choices that define the next chunk

    Parameters
    ----------
    file_descriptor : str | Path | IO
        an open output file descriptor
    """
    title_line: str = ""
    parsed_key: str = ""
    values: dict[str, Any] = {}
    metadata: dict[str, Any] = {}
    next_chunks: list[str | dict[str, Any]] = []

    def __init__(self,
                 file_descriptor: str | Path | IO,
                 parent: Parser  = None,
                 header_line: str = None) -> None:

        super().__init__(file_descriptor, parent)
        self.header_line = header_line
        self.uuid = ''.join(random.choices(string.ascii_letters + string.digits, k=4))
        self._next_parsers = ParsersList(
            self.next_chunks,
            self._available_parsers,
            self.run_metadata
        )

    def _check_for_next(self, line):
        """ Decides on the chunk which a given line of the output should
        belong to.

        Parameters
        ----------
        line : str
            a line of the output file
        Returns
        -------
            None if the line belongs to this chunk; a FileChunk class if it belongs
            to another chunk
        """

        # find the next parser by comparing the line with the given regex
        next_parsers = self._next_parsers.process_line(line)
        for parser_meta in next_parsers:
            line_regex = parser_meta["line"]
            if re.search(line_regex, line):
                return _get_parser(parser_meta.get("parser", None), header_line=line)
        return None

    def collect(self):
        """ Collects all lines belonging to this chunk
        """
        lines = [self.header_line] if self.header_line is not None else []
        while line := self.fd.readline():
            next_chunk = self._check_for_next(line)
            if next_chunk is not None:
                self.next = next_chunk
                break
            lines.append(line)

        # TODO: Here check if the output file is actually complete!
        return "".join(lines)

    def parse(self) -> dict:
        """Generic parsing of the output file chunk"""
        line = self.collect()
        parse_result = {}
        # 1. values parsing
        for k, v in self.values.items():
            if not inspect.isfunction(v):
                v = to_str(v)
            # bring metadata to converter function; useful for parser variables
            match = v(line, self.run_metadata)
            if match is not None:
                parse_result[k] = match  # the first parenthesized subgroup
        # 2. metadata parsing
        for k, v in self.metadata.items():
            if not inspect.isfunction(v):
                v = to_str(v)
            match = v(line, self.run_metadata)
            if match is not None:
                self.parent.run_metadata[k] = match  # the first parenthesized subgroup
        # 3. warnings and errors parsing (ones that begin with optional space and asterisk)
        self.parent.errors += re.findall(r"^(?: ?\*.*\n)+[\s\n]*\Z", line, re.MULTILINE)
        self.parent.warnings += re.findall(r"^(?: ?\*.*\n)+", line, re.MULTILINE)
        self.result = parse_result
        return self.result

    def annotate(self, annotated_file: IO) -> None:
        """Annotates the given chunk"""
        line = self.collect()
        # metadata parsing
        for k, v in self.metadata.items():
            if not inspect.isfunction(v):
                v = to_str(v)
            match = v(line, self.run_metadata)
            if match is not None:
                self.parent.run_metadata[k] = match  # the first parenthesized subgroup
        # get annotation and annotate
        truncated_name = self.name if len(self.name) < 15 else self.name[:12] + "..."
        annotation = f"{truncated_name:>15s}:{self.uuid}"
        annotated_lines = [f"{annotation} -> {el}" for el in line.split("\n")]
        if isinstance(annotated_file, IO):
            annotated_file.write("\n".join(annotated_lines))
        else:
            print("\n".join(annotated_lines), end="")


class ParsersList:
    def __init__(
            self,
            parsers_meta: list[str | dict[str, Any]],
            available_parsers: dict[str, Type[ChunkParser]],
            runtime_choices: dict[str, str]):
        """ A class representing a list of parser metadata

        Parameters
        ----------
        parsers_meta:
            a list of next chunk parsers' metadata
        available_parsers:
            a dictionary of all available parsers
        runtime_choices:

        """
        self._all_parsers = []
        for parser_meta in parsers_meta:
            # parser_meta is just a parser name
            if not isinstance(parser_meta, dict):
                parser_meta = {"chunk": parser_meta}

            parser_name = parser_meta["chunk"]
            if "line_present" in parser_meta:
                # the first string in `chunk` entry is the parser name to be used
                # when the line is present in the chunk; the second is used in the opposite case
                assert isinstance(parser_name, (list, tuple)) and len(parser_name) == 2
                yes_parser, no_parser = parser_name
                next_parser: dict[str, str | bool | dict] = {
                    "line_present": parser_meta["line_present"],
                    "found": False,
                    "parsers": {
                        True: {
                            "line": available_parsers[yes_parser].title_line,
                            "parser": available_parsers[yes_parser],
                        },
                        False: {
                            "line": available_parsers[no_parser].title_line,
                            "parser": available_parsers[no_parser],
                        }
                    }
                }
            else:
                next_parser: dict[str, str | bool | dict] = {
                    "line": available_parsers[parser_name].title_line,
                    "parser": available_parsers[parser_name],
                }
                if "runtime_choices" in parser_meta:
                    next_parser["runtime_choices"] = parser_meta["runtime_choices"]
            self._all_parsers.append(next_parser)
        self._on_parsers = self.process_runtime_choices(runtime_choices)

    def process_runtime_choices(self, runtime_choices):
        """Remove the parsers from the list based on runtime choices of the calculation"""
        # First, throw away all the parsers from all parsers' list based on the runtime choices
        # Second, turn off all the `line_present` parsers
        all_parsers = []
        on_parsers = []
        for parser in self._all_parsers:
            if "line_present" in parser:
                all_parsers.append(parser)
                on_parsers.append(parser["parsers"][False])
                continue

            parser_choices = parser.get("runtime_choices", {})
            on_parser = True
            for choice, value in parser_choices.items():
                # boolean: check only for the existence of the key
                if isinstance(value, bool) and ((choice in runtime_choices) != value):
                    on_parser = False
                elif not isinstance(value, bool) and runtime_choices.get(choice, None) != value:
                    on_parser = False
            if on_parser:
                all_parsers.append(parser)
                on_parsers.append(parser)
        self._all_parsers = all_parsers
        return on_parsers

    def process_line(self, line):
        """Change parsers set based on the line from the output file"""
        on_parsers = []
        for parser in self._all_parsers:
            if "line_present" in parser:
                if not parser["found"]:
                    line_regex = parser["line_present"]
                    parser["found"] = bool(re.search(line_regex, line))
                on_parsers.append(parser["parsers"][parser["found"]])
                continue
            on_parsers.append(parser)
        return on_parsers

    def __iter__(self):
        for parser in self._on_parsers:
            yield parser



"""AIMS output parser base class."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

__author__ = "Thomas A. R. Purcell and Andrey Sobolev"
__version__ = "1.0"
__email__ = "purcellt@arizona.edu and andrey.n.sobolev@gmail.com"
__date__ = "July 2024"

# TARP: Originally an object, but type hinting needs this to be an int
LINE_NOT_FOUND = -1000
EV_PER_A3_TO_KBAR = 1.60217653e-19 * 1e22


class ParseError(Exception):
    """Parse error during reading of a file."""


class AimsParseError(Exception):
    """Exception raised if an error occurs when parsing an Aims output file."""

    def __init__(self, message: str) -> None:
        """Initialize the error with the message, message."""
        self.message = message
        super().__init__(self.message)


@dataclass
class AimsOutSection:
    """Base class for AimsOutSections.

    Attributes:
        _lines (list[str]): The list of all lines in the chunk
        _prop_line_relation (dict[str, list[int]]): Dictionary containing where     
    """

    _lines: list[str] = field(default_factory=list)
    _prop_line_relation: dict[str, list[int]] = field(default_factory=dict)
    _cache: dict[str, Any] = field(default_factory=dict)
    
    def set_prop_line_relations(self, prop_line_keys: dict[str, list[str]]):
        """Setup the property line maps
        
        Args:
            prop_line_keys (dict[str, str]): Dictionary mapping a property to a keystring in the aims.out file
        """
        for key in prop_line_keys.keys():
            if key not in self._prop_line_relation:
                self._prop_line_relation[key] = []
        
        for ll, line in enumerate(self._lines):
            for key, val in prop_line_keys.items():
                if any([vv in line for vv in val]):
                    self._prop_line_relation[key].append(ll)
                    break


    def _parse_key_value_pair(self, key: str, allow_fail=False, dtype: type=str) -> Any:
        """Parse a peice of compilation information
        
        Args:
            key (str): The key to parse
            allow_fail (bool): If True if line is not present return None, else raise an Exception
            dtype (type): Type of the data
        """
        if key in self._cache:
            return self._cache[key]
        
        if key not in self._prop_line_relation:
            return None

        try:
            line_start = self.get_line_inds(key)[-1]
        except IndexError:
            if not allow_fail:
                raise AimsParseError(f"No information about {key} in the aims-output file")
            self._cache[key] = None
            return None
        
        self._cache[key] = dtype(self.lines[line_start].split(":")[1].strip())
        return self._cache[key]
    
    @property 
    def prop_line_relation(self):
        """Get all instences where a particular keyword is found in the section"""
        return self._prop_line_relation

    @property
    def lines(self):
        """The content lines for the calculation"""
        return self._lines
    
    def get_line_inds(self, prop_key: str) -> list[int]:
        """Get all line indexes for particular property key
        
        Args:
            prop_key (str): property key to get the lines for

        Returns:
            list[int]: All lines where that property key is found
        """
        if prop_key not in self._prop_line_relation:
            raise ValueError(f"The property key {prop_key} was not found")

        return self._prop_line_relation[prop_key]



"""A main pyaims class representing Calculation"""
from pyfhiaims.aims_input import AimsInput
from pyfhiaims.control import AimsControlIn
from pyfhiaims.geometry import AimsGeometry
from pyfhiaims.outputs.parser import StdoutParser


class AimsCalc:
    """An Aims calculation"""
    aims_input: AimsInput

    @classmethod
    def from_folder(cls, folder, aims_stdout: str = "aims.out", aims_stderr: str = None):
        """Get Aims Calculation from a folder with files."""




"""Object representing an aims output"""
from dataclasses import dataclass, field

import numpy as np
import json

from pyfhiaims.output_parser.aims_out_calculation_section import AimsOutCalcSection
from pyfhiaims.output_parser.aims_out_header_section import AimsOutHeaderSection
from pyfhiaims.output_parser.aims_out_section import ParseError

from pyfhiaims.control.control import AimsControlIn
from pyfhiaims.geometry.geometry import AimsGeometry

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any

PARSABLE_RESULTS = [
    "energy",
    "free_energy",
    "forces",
    "stress",
    "stresses",
    "magmom",
    "dipole",
    "fermi_energy",
    "n_iter",
    "mulliken_charges",
    "mulliken_spins",
    "hirshfeld_charges",
    "hirshfeld_dipole",
    "hirshfeld_volumes",
    "hirshfeld_atomic_dipoles",
    "dielectric_tensor",
    "polarization",
    "homo",
    "lumo",
    "gap",
    "direct_gap",
]

@dataclass
class AimsCalculation:
    """Object to represent the aims outputs"""

    _geometry: AimsGeometry = None
    _results: dict[str, Any] = field(default=dict)

    def __getitem__(self, key:str) -> Any:
        """Get a particular result"""
        if key not in PARSABLE_RESULTS:
            raise ValueError("Requested item is not a parasable result")
        
        return self._results.get(key)
    
    @property
    def geometry(self):
        return self._geometry

@dataclass
class AimsOutput:
    _images: list[AimsCalculation] = None
    _input_geometry: AimsGeometry = None
    _input_control: AimsControlIn | None = None

    def get_image(self, ind):
        return self._images[ind]
    
    @property
    def get_n_images(self):
        return len(self._images)

    @classmethod
    def from_directory(cls, directory):
        """"""
        with open(f"{directory}/aims.out", "r") as fd:
            aims_out_lines = [line.strip() for line in fd.readlines()]

        with open(f"{directory}/parameters.json", "r") as param_json:
            parameters = json.load(param_json)
            aims_control_in = AimsControlIn.from_dict(parameters)

        return cls.from_aims_out_content(
            aims_out_lines,
            aims_control_in,
        )

    @classmethod
    def from_aims_out_file(cls, aims_outfile, aims_control_in=None):
        """"""
        with open(str(aims_outfile), "r") as fd:
            aims_out_lines = [line.strip() for line in fd.readlines()]

        return cls.from_aims_out_content(aims_out_lines, aims_control_in)
    
    @classmethod
    def from_aims_out_content(cls, aims_out_lines, aims_control_in=None):
        """"""
        header_lines = []
        stopped = False
        # Stop the header once the first SCF cycle begins
        for line in aims_out_lines:
            header_lines.append(line)
            if (
                "Convergence:    q app. |  density  | eigen (eV) | Etot (eV)" in line
                or "Begin self-consistency iteration #" in line
            ):
                stopped = True
                break

        if not stopped:
            raise ParseError("No SCF steps present, calculation failed at setup.")

        header = AimsOutHeaderSection(header_lines)
        if header.is_relaxation:
            section_end_line = "Geometry optimization: Attempting to predict improved coordinates."
        else:
            section_end_line = "Begin self-consistency loop: Re-initialization"

        # If SCF is not converged then do not treat the next chunk_end_line as a
        # new chunk until after the SCF is re-initialized
        ignore_chunk_end_line = False
        line_iter = iter(aims_out_lines)
        aims_sections = []
        while True:
            try:
                line = next(line_iter).strip()  # Raises StopIteration on empty file
            except StopIteration:
                break

            section_lines = []
            while section_end_line not in line or ignore_chunk_end_line:
                section_lines.append(line)
                # If SCF cycle not converged or numerical stresses are requested,
                # don't end chunk on next Re-initialization
                patterns = [
                    ("Self-consistency cycle not yet converged - restarting mixer to attempt better convergence."),
                    (
                        "Components of the stress tensor (for mathematical "
                        "background see comments in numerical_stress.f90)."
                    ),
                    "Calculation of numerical stress completed",
                ]
                if any(pattern in line for pattern in patterns):
                    ignore_chunk_end_line = True
                elif "Begin self-consistency loop: Re-initialization" in line:
                    ignore_chunk_end_line = False

                try:
                    line = next(line_iter).strip()
                except StopIteration:
                    break
            aims_sections.append(AimsOutCalcSection(section_lines, header))

        if header.is_relaxation and any("Final atomic structure:" in line for line in aims_sections[-1].lines):
            aims_sections[-2].lines += aims_sections[-1].lines
            aims_sections = aims_sections[:-1]
        
        aims_calculations = [AimsCalculation(_geometry=sec.geometry, _results=sec.results) for sec in aims_sections]

        return cls(
            _images=aims_calculations,
            _input_geometry=header.input_geometry,
            _input_control=aims_control_in,
        )



from __future__ import annotations

from dataclasses import dataclass, field

import numpy as np

from pyfhiaims.geometry.atom import FHIAimsAtom, ATOMIC_SYMBOLS_TO_NUMBERS
from pyfhiaims.species_defaults.species import SpeciesDefaults

from typing import Sequence

from pathlib import Path

from pyfhiaims.utils.typecast import Vector3D, Matrix3D, to_matrix3d
from typing import Optional, Any


class InvalidGeometryError(Exception):
    """Exception raised if there is a problem with the input geometry"""

    def __init__(self, message: str) -> None:
        """Initialize the error with the message, message"""
        self.message = message
        super().__init__(self.message)


@dataclass
class AimsGeometry:
    """Geometry for the structure to run the calculation for
    
    Parameters
    ----------
    atoms: Sequence[FHIAimsAtom]
        The Atoms for the geometry
    lattice_vectors: Optional[Matrix3D]
        The initial set of lattice vectors for the structure
    lattice_constraints: Sequence[tuple[bool, bool, bool]]
        fix lattice constraints for each nucleus and direction
    species_dict: dict[str, SpeciesDefaults]
        Dictionary to get the species for each symbol in the structure
    hessian_block: Optional[Sequence[tuple[int, int, Matrix3D]]]
        In geometry.in, allows to specify a Hessian matrix explicitly, with one
        line for each 3×3 block.
        The option block consists of nine numbers in column-first
        (Fortran) order. The 3×3 block corresponding to j_atom, i_atom is initialized
        by the transposed of block. The Hessian matrix is input in units of eV/Å2.
    hessian_block_lv: Optional[Sequence[tuple[float, float, Matrix3D]]]
        hessian_block for lattice vectors
    hessian_block_lv_atoms: Optional[Sequence[tuple[float, float, Matrix3D]]]
        hessian_block for degrees of freedom between lattice vectors and atoms
    hessian_file: Optional[bool]
        Indicates that there exists a hessian.aims file to be used to construct 
        the Hessian.
    trust_radius: Optional[float]
        allows to specify the initial trust radius value for the trm 
        relaxation algorithm
    symmetry_n_params: Optional[tuple[int, int]]
        Number of parameters for the lattice and fractional degrees of freedom
    symmetry_params: Optional[Sequence[str]]
        The list of all parametric constraints parameters
    symmetry_lv: Optional[tuple[str, str, str]]
        The list of parametric constraints for the lattice_vectors
    symmetry_frac: Optional[Sequence[tuple[str, str, str]]]
        The list of parametric constraints for the fractional coordinates
    symmetry_frac_change_threshold: Optional[float]
        Specifies the maximum allowed change in the initial structure to any
        fractional coordinate after applying the parametric constraints. If set to 1.0,
        then these tests will be ignored. max_change is the maximum allowed change in 
        the fractional coordinates of an atom
    homogeneous_field: Optional[Vector3D]
        Allows performing a calculation for a system in a homogeneous electrical
        field E in units of V/AA.
    multipole: Optional[Sequence[tuple[float, float, float, int, float]]]
        Places the center of an electrostatic multipole field at a specified location, 
        to simulate an embedding potential.
            x : x coordinate of the multipole.
            y : y coordinate of the multipole.
            z : z coordinate of the multipole.
            order : Integer number, which specifies the order of the multipole
                    (0 or 1 ≡ monopole or dipole).
            charge : Real number, specifies the charge associated with the multipole.
    esp_constraint: Optional[tuple[float, float, float] | tuple[float, float]]
        PBC only. Define the constraints for the fit of the ESP-charges for
        each atom. Depending on the chosen method (`esp_constraint` method in
        control.in). Method 1 needs three parameters (χ, J00, w) 3.167 and method 2
        needs two parameters (q0, β) 3.169 as defined above
    verbatim_writeout: Optional[bool]
        If True write the geometry.in file into the output file
    calculate_friction: Optional[bool]
        Calculate friction for this geometry
    """
    atoms: Sequence[FHIAimsAtom] = None
    lattice_vectors: Optional[Matrix3D[float]] = None
    lattice_constraints: Optional[Vector3D[bool]] = None
    species_dict: Optional[dict[str, SpeciesDefaults]] = field(default_factory=dict)
    hessian_block: Optional[Sequence[tuple[int, int, Matrix3D]]] = None
    hessian_block_lv: Optional[Sequence[tuple[float, float, Matrix3D]]] = None
    hessian_block_lv_atoms: Optional[Sequence[tuple[float, float, Matrix3D]]] = None
    hessian_file: Optional[bool] = None
    trust_radius: Optional[float] = None
    symmetry_n_params: Optional[tuple[int, int, int]] = None
    symmetry_params: Optional[Sequence[str]] = None
    symmetry_lv: Optional[tuple[str, str, str]] = None
    symmetry_frac: Optional[Sequence[tuple[str, str, str]]] = None
    symmetry_frac_change_threshold: Optional[float] = None
    homogeneous_field: Optional[Vector3D] = None
    multipole: Optional[Sequence[tuple[float, float, float, int, float]]] = None
    esp_constraint: Optional[tuple[float, float, float] | tuple[float, float]] = None
    verbatim_writeout: Optional[bool] = None
    calculate_friction: Optional[bool] = None

    def __post_init__(self):
        """Setup some optional inputs given information and verify all inputs are correct"""
        if self.lattice_vectors is not None:
            self.lattice_vectors = np.array(self.lattice_vectors)
            if self.lattice_vectors.shape != (3, 3):
                raise InvalidGeometryError("Lattice vectors must be None or a 3x3 matrix")
            
            if np.linalg.det(self.lattice_vectors) < 1e-12:
                raise InvalidGeometryError("Lattice vectors must be linearly independent")
                
            if self.lattice_constraints is None:
                self.lattice_constraints = np.zeros((3, 3), dtype=bool)            
        else:
            if self.lattice_constraints is not None:
                raise InvalidGeometryError("Lattice vectors must be defined for lattice_constraints to be defined.")
                        
            if any([inp is not None for inp in [
                self.symmetry_n_params,
                self.symmetry_params,
                self.symmetry_lv,
                self.symmetry_frac,
                self.symmetry_frac_change_threshold
            ]]):
                raise InvalidGeometryError("Lattice vectors must be defined when using parametric constraints.")
        self.verify_object_lens()

    def __len__(self):
        """Get the length of the geometry"""
        return len(self.symbols)
    
    def verify_object_lens(self):
        """Verifies the sizes of each input"""            
        for item in ["lattice_constraints", "symmetry_lv"]:
            val = getattr(self, item, None)
            if val is not None and np.array(val).shape != (3, 3):
                raise InvalidGeometryError(f"The shape of {item} must be 3x3 not {np.array(val).shape}")

    @classmethod
    def from_file(cls, filename: Path | str) -> "AimsGeometry":
        """Read in a geometry.in file and construct a Geometry object
        
        Parameters
        ----------
        filename: str | Path
            File to load in

        Returns
        -------
        AimsGeometry
            The Geometry associated with the file
        """
        with open(filename, "r") as fd:
            lines = [line.strip() for line in fd.readlines()]
        
        return cls.from_strings(lines)
        
    @classmethod
    def from_strings(cls, lines: list[str]) -> "AimsGeometry":
        """Read in a geometry.in file and construct a Geometry object
        
        Parameters
        ----------
        lines: list[str]
            a list of all lines in the file.

        Returns
        -------
            The Geometry object associated with the file
        """
        lattice_vectors = []
        symbols: list[str] = []
        coords: list[Vector3D] = []

        is_empty: list[bool] = []
        is_pseudocore: list[bool] = []
        is_fractional: list[bool] = []

        hessian_block: None | list[tuple[int, int, Matrix3D]] = []
        hessian_block_lv: None | list[tuple[float, float, Matrix3D]] = []
        hessian_block_lv_atoms: None | list[tuple[float, float, Matrix3D]] = []

        symmetry_n_params: None | tuple[int, int, int] = (0, 0, 0)
        symmetry_params: None | Sequence[str] = []
        symmetry_lv: None | list[list[str]] = []
        symmetry_frac: None | list[list[str]] = []

        multipole: None |  list[tuple[float, float, float, int, float]] = []

        symmetry_frac_change_threshold: None | float = None
        hessian_file: None | bool = None
        trust_radius: None | float = None
        homogeneous_field: Vector3D = None
        esp_constraint: None | tuple[float, ...] = None
        verbatim_writeout: None | bool = None
        calculate_friction: None | bool = None

        velocities: dict[int, Vector3D] = {}
        initial_charge: dict[int, float] = {}
        initial_moment: dict[int, float] = {}
        nuclear_constraints: dict[int, list[bool]] = {}
        lattice_constraints: None | dict[int, list[bool]] = {}
        constraint_regions: dict[int, int | None] = {}
        magnetic_response: dict[int, bool] = {}
        magnetic_moment: dict[int, float] = {}
        nuclear_spin: dict[int, float] = {}
        isotope: dict[int, float] = {}
        RT_TDDFT_initial_velocity: dict[int, Vector3D] = {}

        last_add: str = ""
        for line in lines:
            if len(line) == 0 or line[0] == "#":
                continue
            inp = line.split("#")[0].split()
        
            if inp[0] in ["atom", "atom_frac", "empty", "pseudocore"]:
                symbols.append(inp[4])
                coords.append([float(ii) for ii in inp[1:4]])
                is_fractional.append(inp[0] == "atom_frac")
                is_empty.append(inp[0] == "empty")
                is_pseudocore.append(inp[0] == "pseudocore")
                last_add = "atom"
            elif inp[0] == "lattice_vector":
                lattice_vectors.append([float(ii) for ii in inp[1:4]])
                last_add = "lattice"
            elif inp[0] == "initial_moment":
                initial_moment[len(coords) - 1] = float(inp[1])
            elif inp[0] == "initial_charge":
                initial_charge[len(coords) - 1] = float(inp[1])
            elif inp[0] == "constrain_relaxation":
                if last_add == "atom":
                    nuclear_constraints[len(coords) - 1] = _create_constraints(inp)
                if last_add == "lattice":
                    lattice_constraints[len(coords) - 1] = _create_constraints(inp)
            elif inp[0] == "velocity":
                velocities[len(coords) - 1] = [float(ii) for ii in inp[1:4]]
            elif inp[0] == "RT_TDDFT_initial_velocity":
                RT_TDDFT_initial_velocity[len(coords) - 1] = [float(ii) for ii in inp[1:4]]
            elif inp[0] == "constraint_region":
                constraint_regions[len(coords) - 1] = int(inp[1])
            elif inp[0] == "magnetic_response":
                magnetic_response[len(coords) - 1] = True
            elif inp[0] == "magnetic_moment":
                magnetic_moment[len(coords) - 1] = float(inp[1])
            elif inp[0] == "nuclear_spin":
                nuclear_spin[len(coords) - 1] = float(inp[1])
            elif inp[0] == "isotope":
                isotope[len(coords) - 1] = float(inp[1])
            elif inp[0] == "symmetry_frac_change_threshold":
                symmetry_frac_change_threshold = float(inp[1])
            elif inp[0] == "hessian_file":
                hessian_file = True
            elif inp[0] == "trust_radius":
                trust_radius = float(inp[1])
            elif inp[0] == "homogeneous_field":
                homogeneous_field = np.array([float(e) for e in inp[1:]])
            elif inp[0] == "multipole":
                multipole.append((float(inp[1]), float(inp[2]), float(inp[3]), int(inp[4]), float(inp[5])))
            elif inp[0] == "esp_constraint":
                esp_constraint = tuple([float(ii) for ii in inp[1:]])
            elif inp[0] == "verbatim_writeout":
                verbatim_writeout = ".true." in inp[1]
            elif inp[0] == "calculate_friction":
                calculate_friction = ".true." in inp[1]
            elif inp[0] == "hessian_block":
                hessian_block.append((int(inp[1]), int(inp[2]), np.array([float(ii) for ii in inp[3:12]]).reshape((3,3))))
            elif inp[0] == "hessian_block_lv":
                hessian_block_lv.append((int(inp[1]), int(inp[2]), np.array([float(ii) for ii in inp[3:12]]).reshape((3,3))))
            elif inp[0] == "hessian_block_lv_atoms":
                hessian_block_lv_atoms.append((int(inp[1]), int(inp[2]), np.array([float(ii) for ii in inp[3:12]]).reshape((3,3))))
            elif inp[0] == "symmetry_n_params":
                symmetry_n_params = (int(inp[1]), int(inp[2]), int(inp[3]))
            elif inp[0] == "symmetry_params":
                symmetry_params = inp[1:]
            elif inp[0] == "symmetry_lv":
                symmetry_lv.append([ii.strip() for ii in line[11:].strip().split(",")])
            elif inp[0] == "symmetry_frac":
                symmetry_frac.append([ii.strip() for ii in line[13:].strip().split(",")])

        def dct2arr(dct: dict[int, Any], n_comp:int) -> np.ndarray[Any, np.dtype[bool]] | None:
            """Convert a dictionary of atom indexes and values to an array"""
            if len(dct) == 0:
                return None
            values = list(dct.values())
            if isinstance(values[0], (Sequence, np.ndarray)):
                to_ret = np.zeros((n_comp, len(values[0])), dtype=type(values[0]))
            else:
                to_ret = np.zeros(n_comp, dtype=type(values[0]))
            for key, val in dct.items():
                to_ret[key] = val
            return to_ret
        
        if len(lattice_vectors) > 0:
            try:
                lattice_vectors = to_matrix3d(lattice_vectors)
            except AssertionError:
                raise InvalidGeometryError("There must be 3 lattice vectors")

            lattice_constraints = dct2arr(lattice_constraints, 3)

        for cc, coord in enumerate(coords):
            if is_fractional[cc]:
                coords[cc] = list(np.dot(coord, lattice_vectors))
        
        atoms = []
        for aa, sym in enumerate(symbols):
            atoms.append(
                FHIAimsAtom(
                    symbol=sym,
                    position=coords[aa],
                    velocity=velocities.get(aa),
                    initial_charge=initial_charge.get(aa, 0.0),
                    initial_moment=initial_moment.get(aa, 0.0),
                    constraints=nuclear_constraints.get(aa),
                    constraint_region=constraint_regions.get(aa),
                    magnetic_response=magnetic_response.get(aa),
                    magnetic_moment=magnetic_moment.get(aa),
                    nuclear_spin=nuclear_spin.get(aa),
                    isotope=isotope.get(aa),
                    is_empty=is_empty[aa],
                    is_pseudocore=is_pseudocore[aa],
                    RT_TDDFT_initial_velocity=RT_TDDFT_initial_velocity.get(aa),
                )
            )

        if len(lattice_vectors) > 0:
            for aa in range(len(atoms)):
                atoms[aa].set_fractional(lattice_vectors)

        if symmetry_n_params == (0, 0, 0):
            symmetry_lv = None
            symmetry_frac = None
            symmetry_params = None
            symmetry_n_params = None
            symmetry_frac_change_threshold = None

        if len(lattice_vectors) == 0:
            lattice_vectors = None
            lattice_constraints = None

        if len(hessian_block) == 0:
            hessian_block = None
        if len(hessian_block_lv) == 0:
            hessian_block_lv = None
        if len(hessian_block_lv_atoms) == 0:
            hessian_block_lv_atoms = None
        if len(multipole) == 0:
            multipole = None

        return cls(
            atoms=atoms,
            lattice_vectors=lattice_vectors,
            lattice_constraints=lattice_constraints,
            hessian_block=hessian_block,
            hessian_block_lv=hessian_block_lv,
            hessian_block_lv_atoms=hessian_block_lv_atoms,
            hessian_file=hessian_file,
            trust_radius=trust_radius,
            symmetry_n_params=symmetry_n_params,
            symmetry_params=symmetry_params,
            symmetry_lv=symmetry_lv,
            symmetry_frac=symmetry_frac,
            symmetry_frac_change_threshold=symmetry_frac_change_threshold,
            homogeneous_field=homogeneous_field,
            multipole=multipole,
            esp_constraint=esp_constraint,
            verbatim_writeout=verbatim_writeout,
            calculate_friction=calculate_friction,
        )

    def to_atoms(self):
        """Convert AimsGeometry to an ase.Atoms object."""
        from ase import Atoms
        atoms = Atoms(
            symbols=self.symbols,
            positions=self.positions,
            magmoms=self.magnetic_moments,
            charges=self.initial_charges,
            pbc=self.lattice_vectors is not None,
            cell=self.lattice_vectors if self.lattice_vectors is not None else (0, 0, 0),
        )
        return atoms


    def load_species(self, species_directory:str | Path):
        """Create a species dictionary for the atoms
        
        Parameters
        ----------
        species_directory:str | Path
            The Path to load the species into
        """
        for sym in np.unique([at.symbol for at in self.atoms]):
            number = ATOMIC_SYMBOLS_TO_NUMBERS[sym]
            if number < 100:
                self.species_dict[sym] = SpeciesDefaults.from_file(f"{species_directory}/{number:02d}_{sym}_default")
            else:
                self.species_dict[sym] = SpeciesDefaults.from_file(f"{species_directory}/{number:03d}_{sym}_default")

    def set_species(self, sym: str, species: SpeciesDefaults):
        """Sets a species default for a given symbol
        
        Parameters
        ----------
        sym: str
            The symbol to add the species for
        species: SpeciesDefaults
            The species for the symbol
        """
        self.species_dict[sym] = species

    def get_species(self, sym: str) -> SpeciesDefaults:
        """Gets a species default for a given symbol
        
        Parameters
        ----------
        sym: str
            The symbol to add the species for
        
        Returns
        -------
            The species for the symbol
        """
        return self.species_dict[sym]
        
    @property
    def file_content(self):
        """Get the file content for this geometry"""
        content_str = ["# Making from pyaims", ]
        if self.lattice_vectors is not None and 3 == len(self.lattice_vectors):
            for lv, lv_const in zip(self.lattice_vectors, self.lattice_constraints):
                content_str.append(f"lattice_vector {lv[0]:.15e} {lv[1]:.15e} {lv[2]:.15e}")
                if np.any(lv_const):
                    content_str.append(f"    constrain_relaxation "
                                       f"{'x ' if lv_const[0] else ''}"
                                       f"{'y ' if lv_const[1] else ''}"
                                       f"{'z' if lv_const[2] else ''}"
                                       )
        for atom in self.atoms:
            content_str.append(atom.to_string)

        if self.homogeneous_field is not None:
            if len(self.homogeneous_field) != 3:
                raise InvalidGeometryError("The provided homogeneous_field value is invalid")
            content_str.append(f"homogeneous_field {self.homogeneous_field[0]:>20.12e} "
                               f"{self.homogeneous_field[1]:>20.12e} {self.homogeneous_field[2]:>20.12e}")

        if self.multipole is not None:
            if len(self.multipole) != 5:
                raise InvalidGeometryError("The provided multipole value is invalid")
            content_str.append(f"multipole {self.multipole[0]:>20.12e} {self.multipole[1]:>20.12e} "
                               f"{self.multipole[2]:>20.12e} {self.multipole[3]} "
                               f"{self.multipole[4]:>20.12e}")
            # TODO: Specify a dipole moment to multipoles

        if self.esp_constraint is not None:
            if len(self.esp_constraint) not in (2, 3):
                raise InvalidGeometryError("The provided esp_constraint value is invalid")
            content_str.append(f"esp_constraint {' '.join(f'{x:>20.12e}' for x in self.esp_constraint)}")

        if self.verbatim_writeout is not None:
            content_str.append(f"verbatim_writeout {'.true.' if self.verbatim_writeout else '.false.'}")

        if self.calculate_friction is not None:
            content_str.append(f"calculate_friction {'.true.' if self.calculate_friction else '.false.'}")

        if self.symmetry_n_params:
            content_str.append(f"symmetry_n_params {self.symmetry_n_params[0]} {self.symmetry_n_params[1]} "
                            f"{self.symmetry_n_params[2]}")
 
        if self.symmetry_params:
            content_str.append(f"symmetry_params " + " ".join(self.symmetry_params))
        if self.symmetry_lv is not None:
            for line in self.symmetry_lv:
                content_str.append(f"symmetry_lv " + " , ".join(line))
        if self.symmetry_frac is not None:
            for line in self.symmetry_frac:
                content_str.append(f"symmetry_frac " + " , ".join(line))
        if self.symmetry_frac_change_threshold:
            content_str.append(f"symmetry_frac_change_threshold {self.symmetry_frac_change_threshold}")

        if self.hessian_block is not None:
            for line in self.hessian_block:
                content_str.append(f"hessian_block {line[0]} {line[1]} " +
                                   " ".join([f"{ll:>20.12e}" for ll in line[2:]]))
        if self.hessian_block_lv is not None:
            for line in self.hessian_block_lv:
                content_str.append(f"hessian_block_lv {line[0]} {line[1]} " +
                                   " ".join([f"{ll:>20.12e}" for ll in line[2:]]))
        if self.hessian_block_lv_atoms is not None:
            for line in self.hessian_block_lv_atoms:
                content_str.append(f"hessian_block_lv_atoms {line[0]} {line[1]} " +
                                   " ".join([f"{ll:>20.12e}" for ll in line[2:]]))
        
        if self.hessian_file:
            content_str.append("hessian_file")
        if self.trust_radius is not None:
            content_str.append(f"trust_radius {self.trust_radius}")
        return "\n".join(content_str)

    @property
    def symbols(self):
        return [at.symbol for at in self.atoms]
    
    @property
    def numbers(self):
        return [at.number for at in self.atoms]
    
    @property
    def positions(self):
        return [at.position for at in self.atoms]
    
    @property
    def fractional_positions(self):
        return [at.fractional_position for at in self.atoms]
    
    @property
    def velocities(self):
        return [at.velocity for at in self.atoms]
    
    @property
    def RT_TDDFT_initial_velocities(self):
        return [at.RT_TDDFT_initial_velocity for at in self.atoms]
    
    @property
    def initial_charges(self):
        return [at.initial_charge for at in self.atoms]
    
    @property
    def initial_moments(self):
        return [at.initial_moment for at in self.atoms]
    
    @property
    def nuclear_constraints(self):
        return [at.constraints for at in self.atoms]
    
    @property
    def constraint_regions(self):
        return [at.constraint_region for at in self.atoms]
    
    @property
    def is_empty_atoms(self):
        return [at.is_empty for at in self.atoms]
    
    @property
    def is_pseudocore_atoms(self):
        return [at.is_pseudocore for at in self.atoms]
    
    @property
    def magnetic_responses(self):
        return [at.magnetic_response for at in self.atoms]
    
    @property
    def magnetic_moments(self):
        return [at.magnetic_moment for at in self.atoms]
    
    @property
    def nuclear_spins(self):
        return [at.nuclear_spin for at in self.atoms]
    
    @property
    def isotopes(self):
        return [at.isotope for at in self.atoms]

    @property
    def species_block(self):
        """Get the species block for the control.in file"""
        if any([sym not in self.species_dict for sym in np.unique(self.symbols)]):
            raise InvalidGeometryError("Species are not defined for all atoms in the structure")
        return "\n".join([self.species_dict[sym].content for sym in np.unique(self.symbols)])

    @property
    def n_atoms(self):
        """The number of atoms in the structure"""
        return len(self.symbols)
    
    @property
    def masses(self) -> np.ndarray:
        """Get the masses of the structure"""
        return self._species_property("mass")

    @property
    def nuclear_charges(self) -> np.ndarray:
        """Gets the nuclear charges of the structure"""
        return self._species_property("nucleus")

    def _species_property(self, prop: str) -> np.ndarray:
        """Returns an array with property values corresponding to atomic Species."""
        if any([sym not in self.species_dict for sym in np.unique(self.symbols)]):
            raise InvalidGeometryError("Species are not defined for all atoms in the structure")
        return np.array([getattr(self.species_dict[sym], prop) * (not self.is_empty_atoms[ss])
                         for ss, sym in enumerate(self.symbols)])


def _create_constraints(line: list[str]) -> list[bool]:
    """Sets constraints"""
    constraints = [False, False, False]
    if "x" in line[1:]:
        constraints[0] = True
    if "y" in line[1:]:
        constraints[1] = True
    if "z" in line[1:]:
        constraints[2] = True
    if "true" in line[1:]:
        constraints[:] = True
    return constraints



""" Type converters for parsed values
"""
import re
import datetime
import tempfile
from pathlib import Path



def _to_builtin(pattern, builtin_type):
    """Converter to the builtin type"""
    def _helper(line, _):
        match = re.search(pattern, line, re.DOTALL | re.MULTILINE)
        if match:
            return builtin_type(match.group(1))
        return None
    return _helper


def to_atoms(pattern, from_input=False):
    """Convert aims.out geometry representation to ase.Atoms"""
    from ase.io import read

    def _helper(line, _):
        # Capture everything to the next dashed line including it
        match = re.search(rf"(?:{pattern}).*\n([\s\S]*?)\n *-{{60,}}", line, re.MULTILINE)
        if match is None:
            return None
        lines = match.group()
        # remove fractional coordinates
        lines = lines[:lines.find("Fractional coordinates")]
        lines = lines.split("\n")[2:-1]
        with tempfile.TemporaryDirectory() as tmp:
            file_name = Path(tmp) / "geometry.in"
            with open(file_name, "w") as f:
                f.write("\n".join(lines))
            atoms = read(file_name, format="aims")
        return atoms
    return _helper


def to_bool(pattern):
    def _helper(line, _):
        return bool(re.search(pattern, line, re.DOTALL | re.MULTILINE))
    return _helper


def to_float(pattern):
    return _to_builtin(pattern, float)


def to_int(pattern):
    return _to_builtin(pattern, int)


def to_str(pattern):
    return _to_builtin(pattern, str)


def to_date(pattern):
    return _to_builtin(pattern, lambda x: datetime.datetime.strptime(x, '%Y%m%d').date())


def to_time(pattern):
    return _to_builtin(pattern, lambda x: datetime.datetime.strptime(x, '%H%M%S.%f').time())


def to_table(pattern, *, num_rows: int | str, header: int = 1, dtype: list[type] | tuple[type] = None):
    """Convert `num_rows` lines after `header` indicated by `pattern` to Python 2D list"""
    dtype = dtype if dtype is not None else []
    def _helper(line, metadata):
        # we need a local scope variable for not to lose access to num_rows
        n_rows = num_rows if isinstance(num_rows, int) else metadata[num_rows]
        match = re.search(rf"{pattern}\n((?:.*\n){{{header+n_rows-1}}})", line, re.MULTILINE)
        if match is None:
            return None
        table = match.group(1).split("\n")[header-1:-1]
        types = dtype + ([None] * (len(table[0].split()) - len(dtype)))
        result = []
        for line in table:
            result.append([t(v) for t, v in zip(types, line.split()) if t is not None])
        # make it 1D list if only one column is asked for in the parser
        if len(result[0]) == 1:
            result = [r[0] for r in result]
        return result
    return _helper


def to_vector(pattern, *, dtype: type = float, multistring: bool = False):
    """Convert a set of numbers to a 1D numpy array.

    Args:
          pattern (str): A regular expression pattern.
          dtype (type): A Python type to convert to.
          multistring (bool): If True, checks for different occurrences of the pattern in a line
          (different matches).
          If False, finds the numbers written in one line (different groups within one match).
    """
    def _helper(line, _):
        match = re.findall(pattern, line, re.DOTALL | re.MULTILINE)
        if not match:
            return None
        return [dtype(x) for x in match] if multistring else [dtype(x) for x in match[0]]
    return _helper


"""Tests for basis_function module."""

from pyfhiaims.species_defaults.basis_function import (
    BasisFunction,
    HydroBasisFunction,
    IonicBasisFunction, GaussianBasisFunction
)

def test_basis_function():
    """Tests a generic BasisFunction constructor."""
    string = "    hydro 4 f 5.6"
    basis_function = BasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert basis_function.enabled
    string = "#    hydro 4 f 5.6"
    basis_function = BasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert not basis_function.enabled


def test_hydro_basis_function():
    """Tests hydrogen-like BasisFunction constructor."""
    string = "    hydro 4 f 5.6"
    basis_function = HydroBasisFunction.from_string(string)
    assert basis_function.type == "hydro"
    assert basis_function.z_eff == 5.6


def test_ionic_basis_function():
    """Tests ionic BasisFunction constructor."""
    string = "    ionic 4 d auto"
    basis_function = IonicBasisFunction.from_string(string)
    assert basis_function.n == 4
    assert basis_function.radius == "auto"
    string = "    ionic 3 d 2.5"
    basis_function = IonicBasisFunction.from_string(string)
    assert basis_function.n == 3
    assert basis_function.radius == 2.5


def test_gaussian_basis_function():
    """Tests Gaussian-based BasisFunction constructor."""
    string = "gaussian 0 1  0.105555E+08"
    basis_function = GaussianBasisFunction.from_string(string)
    assert basis_function.enabled
    assert basis_function.type == "gaussian"
    assert basis_function.n == 1
    assert len(basis_function.alpha_i) == 1
    assert basis_function.alpha_i[0] == 1.05555e7
    string = "# gaussian 0 1  0.105555E+08"
    basis_function = BasisFunction.from_string(string)
    assert not basis_function.enabled



"""An object representing parse results of FHI-aims standard output."""
import os.path
from pathlib import Path
from typing import IO, Any
from pyfhiaims.outputs.parser import StdoutParser


class AimsStdout:

    def __init__(self, stdout_file: str | Path | IO):
        if isinstance(stdout_file, str | Path):
            self.file_name = Path(stdout_file).resolve().as_posix()
            file_path = Path(stdout_file)
            if not file_path.is_file():
                raise FileNotFoundError(f"FHI-aims output file {self.file_name} "
                                        f"does not exist")
        else:
            self.file_name = os.path.realpath(stdout_file.name)
        self._parser = StdoutParser(stdout_file)
        self._results = self._parser.parse()

    @property
    def results(self) -> dict[str, Any]:
        """A dictionary with results for FHI-aims calculation."""
        return self._results

    @property
    def metadata(self) -> dict[str, Any]:
        """A metadata dictionary for FHI-aims calculation, including runtime choices and
        some background calculation checks."""
        return self._parser.run_metadata

    @property
    def warnings(self) -> list[str]:
        """A list of warning messages for FHI-aims calculation.
        A warning message is the one beginning with an optional space and one or several
        asterisks."""
        return self._parser.warnings

    @property
    def errors(self) -> list[str]:
        """A list of error messages for FHI-aims calculation.
        An error message is a warning at the end of the file."""
        return self._parser.errors

    @property
    def is_finished_ok(self) -> bool:
        """A check if the calculation is finished successfully."""
        return self._results.get("is_finished_ok", False)

    @property
    def energy(self) -> float:
        """Energy value given after the calculation has finished"""
        if not self.is_finished_ok:
            raise ValueError("Calculation has not finished successfully")
        return self._results["final"]["energy"]

    @property
    def band_gap(self) -> float:
        """Band gap value from the last SCF step."""
        if not self.is_finished_ok:
            raise ValueError("Calculation has not finished successfully")
        return self._results["ionic_steps"][-1]["scf_steps"][-1]["gap"]

    @property
    def forces(self) -> list[float]:
        """Forces from the last ionic step if present."""
        return self._results["ionic_steps"][-1].get("atomic_forces", None)

    def __getattr__(self, item):
        """A shortcut to getting results."""
        if item in self._results:
            return self._results[item]
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{item}'")


"""Test new FHI-aims standard output parser."""

import yaml

from pyfhiaims.outputs.parser import StdoutParser
from tests.utils import is_subset


def test_parser(data_dir):
    """Checks parsing given output files against reference values."""
    stdout_data_dir = data_dir / "stdout"
    for file_name in stdout_data_dir.iterdir():
        if file_name.is_file():
            parser = StdoutParser(file_name)
            results = parser.parse()
            with open(stdout_data_dir / "ref" / (file_name.stem + ".yaml"), "r") as f:
                ref_data = yaml.safe_load(f)
            assert is_subset(ref_data, results)


# def test_print_parsed_results(data_dir):
#     """This test is for development purposes only."""
#     file_name = data_dir / "stdout" / "relax.out.gz"
#     parser = StdoutParser(file_name)
#     results = parser.parse()
#     from pprint import pprint
#     pprint(results)


"""Electronic configurations for the FHI-aims species' defaults"""

from dataclasses import dataclass, field
from enum import Enum

from pyfhiaims.control.chunk import AimsControlChunk


class Orbital(Enum):
    S = 0
    P = 1
    D = 2
    F = 3
    G = 4
    H = 5
    I = 6


class ElectronicConfigurationType(Enum):
    VALENCE = "valence"
    IONIC = "ion_occ"


@dataclass
class ElectronicConfiguration(AimsControlChunk):
    """Electronic configuration, both valence and ionic"""
    type: ElectronicConfigurationType = None
    occupations: dict[(int, Orbital): int] = field(default_factory=dict)

    @classmethod
    def from_strings(cls, config_str) -> "ElectronicConfiguration":
        """Get the electronic configuration from a multi-line string"""
        config_type = None
        config = {}
        for line in config_str:
            type_str, n, l, occ = line.split()
            if config_type is None:
                config_type = ElectronicConfigurationType(type_str)
            else:
                assert config_type == ElectronicConfigurationType(type_str)
            config[(int(n), Orbital[l.upper()])] = int(float(occ))
        return cls(
            type=config_type,
            occupations=config,
        )

    @property
    def n_electrons(self) -> int:
        """The number of electrons in the given configuration"""
        num = 0
        for k, v in self.occupations.items():
            n, l = k
            num += (n - l.value - 1) * (4 * l.value + 2) + v
        return num

    def to_string(self) -> str:
        """A string representation of the configuration"""
        return "\n".join([f"{self.type.value:>11} {k[0]:6d}  {k[1].name.lower()} {v:3d}."
                          for k, v in self.occupations.items()])


class ValenceElectronicConfiguration(ElectronicConfiguration):
    type = ElectronicConfigurationType.VALENCE
    keywords = ("valence", )


class IonicElectronicConfiguration(ElectronicConfiguration):
    type = ElectronicConfigurationType.IONIC
    keywords = ("ion_occ", )


"""Classes for reading/manipulating/writing FHI-aims control.in files."""
from __future__ import annotations

import re
import time
from copy import deepcopy
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn

from monty.json import MontyDecoder, MSONable

from pyfhiaims.control.kpoints import AimsKPoints
from pyfhiaims.errors import PyaimsError
from pyfhiaims.geometry.geometry import AimsGeometry
from pyfhiaims.species_defaults.species import SpeciesDefaults

if TYPE_CHECKING:
    from typing import Any
    from typing_extensions import Self

__author__ = "Thomas A. R. Purcell"
__version__ = "1.0"
__email__ = "purcellt@arizona.edu"
__date__ = "July 2024"


@dataclass
class AimsControlIn(MSONable):
    """An FHI-aims control.in file.

    Attributes:
        parameters (dict[str, Any]): The parameters' dictionary containing all input
            flags (key) and values for the control.in file
    """

    parameters: dict[str, Any] = field(default_factory=dict)
    outputs: list[str] = field(default_factory=list)
    k_points: AimsKPoints = None
    species_defaults: dict[str, SpeciesDefaults] = field(default_factory=dict)  # None?

    def __getitem__(self, key: str) -> Any:
        """Get an input parameter.

        Args:
            key (str): The parameter to get

        Returns:
            The setting for that parameter

        Raises:
            KeyError: If the key is not in self._parameters
        """
        if key == "output":
            return self.outputs

        if key not in self.parameters:
            raise KeyError(f"{key} not set in AimsControlIn")
        return self.parameters[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Set an attribute of the class.

        Args:
            key (str): The parameter to get
            value (Any): The value for that parameter
        """
        if key == "output":
            warn(
                "Outputs are set seperately, use the outputs property", 
                RuntimeWarning, 
                stacklevel=1
            )
        else:
            self.parameters[key] = value

    def __delitem__(self, key: str) -> Any:
        """Delete a parameter from the input object.

        Args:
        key (str): The key in the parameter to remove

        Returns:
            Either the value of the deleted parameter or None if key is
            not in self._parameters
        """
        if key == "output":
            self.outputs = []

        return self.parameters.pop(key, None)

    @staticmethod
    def get_aims_control_parameter_str(key: str, value: Any, fmt: str) -> str:
        """Get the string needed to add a parameter to the control.in file.

        Args:
            key (str): The name of the input flag
            value (Any): The value to be set for the flag
            fmt (str): The format string to apply to the value

        Returns:
            str: The line to add to the control.in file
        """
        if value is None:
            return ""
        return f"{key:50s}{fmt % value}\n"

    def get_content(
        self, geometry: AimsGeometry, verbose_header: bool = False, directory: str | Path | None = None
    ) -> str:
        """Get the content of the file.

        Args:
            geometry (AimsGeometry): The geometry to write the input file for
            verbose_header (bool): If True print the input option dictionary
            directory: str | Path | None = The directory for the calculation,

        Returns:
            str: The content of the file for a given geometry
        """
        parameters = deepcopy(self.parameters)

        if directory is None:
            directory = ""

        lim = "#" + "=" * 79
        content = ""

        if parameters["xc"] == "LDA":
            parameters["xc"] = "pw-lda"

        cubes = parameters.pop("cubes", [])

        if verbose_header:
            content += "# \n# List of parameters used to initialize the calculator:"
            for param, val in parameters.items():
                content += f"#     {param}:{val}\n"
        content += f"{lim}\n"

        if all([inp in parameters for inp in ["smearing", "occupation_type"]]):
            raise ValueError("Both smearing and occupation_type can't be in the same parameters file.")

        for key, value in parameters.items():
            if key in ["species_dir", "plus_u"]:
                continue
            if key == "smearing":
                name = parameters["smearing"][0].lower()
                if name == "fermi-dirac":
                    name = "fermi"
                width = parameters["smearing"][1]
                if name == "methfessel-paxton":
                    order = parameters["smearing"][2]
                    order = " %d" % order
                else:
                    order = ""

                content += self.get_aims_control_parameter_str("occupation_type", (name, width, order), "%s %f%s")
            elif key == "vdw_correction_hirshfeld" and value:
                content += self.get_aims_control_parameter_str(key, "", "%s")
            elif isinstance(value, bool):
                content += self.get_aims_control_parameter_str(key, str(value).lower(), ".%s.")
            elif isinstance(value, (tuple, list)):
                content += self.get_aims_control_parameter_str(key, " ".join(map(str, value)), "%s")
            elif isinstance(value, str):
                content += self.get_aims_control_parameter_str(key, value, "%s")
            else:
                content += self.get_aims_control_parameter_str(key, value, "%r")

        for output_type in self.outputs:
            content += self.get_aims_control_parameter_str("output", output_type, "%s")

        if cubes:
            for cube in cubes:
                content += cube.control_block

        content += f"{lim}\n\n"
        
        for sp in geometry.species_dict.values():
            content += sp.content

        return content

    def write_file(
        self,
        geometry: AimsGeometry,
        directory: str | Path | None = None,
        verbose_header: bool = False,
        overwrite: bool = False,
    ) -> None:
        """Write the control.in file.

        Args:
            geometry (AimsGeometry): The structure to write the input
                file for
            directory (str or Path): The directory to write the control.in file.
                If None use cwd
            verbose_header (bool): If True print the input option dictionary
            overwrite (bool): If True allow to overwrite existing files

        Raises:
            ValueError: If a file must be overwritten and overwrite is False
            ValueError: If k-grid is not provided for the periodic structures
        """
        directory = directory or Path.cwd()

        if (Path(directory) / "control.in").exists() and not overwrite:
            raise ValueError(f"control.in file already in {directory}")

        if (geometry.lattice_vectors is not None) and (
            "k_grid" not in self.parameters and "k_grid_density" not in self.parameters
        ):
            raise ValueError("k-grid must be defined for periodic systems")

        content = self.get_content(geometry, verbose_header)

        with open(f"{directory}/control.in", mode="w") as file:
            file.write(f"#{'=' * 72}\n")
            file.write(f"# FHI-aims geometry file: {directory}/geometry.in\n")
            file.write("# File generated from pyaims\n")
            file.write(f"# {time.asctime()}\n")
            file.write(f"#{'=' * 72}\n")

            file.write(content)

    def as_dict(self) -> dict[str, Any]:
        """Get a dictionary representation of the geometry.in file."""
        dct: dict[str, Any] = {
            "@module": type(self).__module__,
            "@class": type(self).__name__,
            "parameters": self.parameters,
            "outputs": self.outputs
        }
        return dct

    @classmethod
    def from_file(cls, control_file: str | Path) -> "AimsControlIn":
        """Instantiate the Control object """
        with open(control_file, "r") as f:
            lines = f.read()
        parameters = {}
        outputs = []

        # get species' defaults from the file first
        species = {}
        species_re = re.compile(r"(?<=\n) *species.*?(?=\n *species|$)", re.DOTALL)
        species_lines = re.findall(species_re, lines)
        element_re = re.compile(r" *species *(\S+)", re.DOTALL)

        for block in species_lines:
            if re.match(element_re, block) is not None:
                element = re.match(element_re, block).group(1)
                species[element] = SpeciesDefaults.from_strings(block.split("\n"))

        # then everything else
        for line in lines.split("\n"):
            # remove comments and blank lines
            idx = line.find("#")
            line = line.strip() if idx == -1 else line[:idx].strip()
            if not line:
                continue
            # stop at species
            if line.startswith("species "):
                break
            k, v = line.split(maxsplit=1)
            if k == "output":
                if "cube" in v:
                    # TODO: does not work with Cubes yet
                    raise PyaimsError("Reading control.in from file does not work with cubes yet")
                outputs.append(v)
            else:
                parameters[k] = v
        return AimsControlIn(
            parameters=parameters,
            outputs=outputs,
            species_defaults=species
        )

    @classmethod
    def from_dict(cls, dct: dict[str, Any]) -> Self:
        """Initialize from dictionary.

        Args:
            dct (dict[str, Any]): The MontyEncoded dictionary

        Returns:
            The AimsControlIn for dct
        """
        decoded = {key: MontyDecoder().process_decoded(val) for key, val in dct.items() if not key.startswith("@")}

        return cls(parameters=decoded["parameters"], outputs=decoded["outputs"])

